
from typing import (Any, Callable, Collection, DefaultDict,
                    List, Optional,
                    TYPE_CHECKING)

from glycresoft.chromatogram_tree.chromatogram import GlycopeptideChromatogram
from glycresoft.chromatogram_tree.mass_shift import MassShiftBase
from glycresoft.chromatogram_tree import (
    ChromatogramWrapper,
    Unmodified)

from glycresoft.task import log_handle

from ..spectrum_match import SpectrumMatch, SpectrumSolutionSet

from .base import SolutionEntry, Predicate, always, TargetType
from .representer import RepresenterSelectionStrategy, TotalBestRepresenterStrategy


class SpectrumMatchSolutionCollectionBase(object):
    tandem_solutions: List[SpectrumSolutionSet]
    best_msms_score: float
    representative_solutions: List[SolutionEntry]

    def compute_representative_weights(self, threshold_fn: Predicate = always, reject_shifted: bool = False,
                                       targets_ignored: Collection = None,
                                       strategy: Optional[RepresenterSelectionStrategy] = None, **kwargs) -> List[SolutionEntry]:
        """Calculate a total score for all matched structures across all time points for this
        solution collection, and rank them.

        This total score is the sum of the score over all spectrum matches for which that
        structure was the best match. The percentile is the ratio of the total score for the
        `i`th structure divided by the sum of total scores over all structures.

        Parameters
        ----------
        threshold_fn: Callable
            A function that filters out invalid solutions based on some criteria, e.g.
            not passing the FDR threshold.
        reject_shifted: bool
            Whether or not to omit any solution that was not mass-shifted. Defaults to False

        Returns
        -------
        list
            A list of solutions, ranked by percentile.
        """
        if strategy is None:
            strategy = TotalBestRepresenterStrategy()
        weights = strategy(self, threshold_fn=threshold_fn,
                           reject_shifted=reject_shifted, targets_ignored=targets_ignored, **kwargs)
        return weights

    def filter_entries(self, entries: List[SolutionEntry], percentile_threshold: float = 1e-5) -> List[SolutionEntry]:
        # This difference is not using the absolute value to allow for scenarios where
        # a worse percentile is located at position 0 e.g. when hoisting via parsimony.
        representers = [x for x in entries if (
            entries[0].percentile - x.percentile) < percentile_threshold]
        return representers

    def most_representative_solutions(self, threshold_fn: Predicate = always, reject_shifted: bool = False,
                                      targets_ignored: Collection = None,
                                      percentile_threshold: float = 1e-5) -> List[SolutionEntry]:
        """Find the most representative solutions, the (very nearly the same, hopefully) structures with
        the highest aggregated score across all MSn events assigned to this collection.

        Parameters
        ----------
        threshold_fn: Callable
            A function that filters out invalid solutions based on some criteria, e.g.
            not passing the FDR threshold.
        reject_shifted: bool
            Whether or not to omit any solution that was not mass-shifted. Defaults to False
        percentile_threshold : float, optional
            The difference between the worst and best percentile to be reported. Defaults to 1e-5.

        Returns
        -------
        list
            A list of solutions with approximately the greatest weight
        """
        weights = self.compute_representative_weights(
            threshold_fn, reject_shifted=reject_shifted, targets_ignored=targets_ignored)
        if weights:
            representers = self.filter_entries(
                weights, percentile_threshold=percentile_threshold)
            return representers
        else:
            return []

    def solutions_for(self, structure: TargetType, threshold_fn: Predicate = always,
                      reject_shifted: bool = False) -> List[SpectrumMatch]:
        '''Get all spectrum matches in this collection for a given
        structure.

        Parameters
        ----------
        structure : Hashable
            The structure collect matches for.
        threshold_fn: Callable
            A function that filters out invalid solutions based on some criteria, e.g.
            not passing the FDR threshold.
        reject_shifted: bool
            Whether or not to omit any solution that was not mass-shifted. Defaults to False

        Returns
        -------
        list
        '''
        solutions = []
        for sset in self.tandem_solutions:
            try:
                psm = sset.solution_for(structure)
                if threshold_fn(psm):
                    if psm.mass_shift != Unmodified and reject_shifted:
                        continue
                    solutions.append(psm)
            except KeyError:
                continue
        return solutions

    def best_match_for(self, structure: TargetType, threshold_fn: Predicate = always) -> SpectrumMatch:
        solutions = self.solutions_for(structure, threshold_fn=threshold_fn)
        if not solutions:
            raise KeyError(structure)
        best_score = -float('inf')
        best_match = None
        for sol in solutions:
            if best_score < sol.score:
                best_score = sol.score
                best_match = sol
        return best_match

    def has_scan(self, scan_id: str) -> bool:
        return any([sset.scan_id == scan_id for sset in self.tandem_solutions])

    def get_scan(self, scan_id: str) -> SpectrumMatch:
        for sset in self.tandem_solutions:
            if sset.scan_id == scan_id:
                return sset
        raise KeyError(scan_id)


class TandemAnnotatedChromatogram(ChromatogramWrapper, SpectrumMatchSolutionCollectionBase):
    time_displaced_assignments: List[SpectrumSolutionSet]

    def __init__(self, chromatogram):
        super(TandemAnnotatedChromatogram, self).__init__(chromatogram)
        self.tandem_solutions = []
        self.time_displaced_assignments = []
        self.best_msms_score = None
        self.representative_solutions = None

    def bisect_charge(self, charge):
        new_charge, new_no_charge = map(
            self.__class__, self.chromatogram.bisect_charge(charge))
        for hit in self.tandem_solutions:
            if hit.precursor_information.charge == charge:
                new_charge.add_solution(hit)
            else:
                new_no_charge.add_solution(hit)
        return new_charge, new_no_charge

    def bisect_mass_shift(self, mass_shift):
        new_mass_shift, new_no_mass_shift = map(
            self.__class__, self.chromatogram.bisect_mass_shift(mass_shift))
        for hit in self.tandem_solutions:
            if hit.best_solution().mass_shift == mass_shift:
                new_mass_shift.add_solution(hit)
            else:
                new_no_mass_shift.add_solution(hit)
        return new_mass_shift, new_no_mass_shift

    def split_sparse(self, delta_rt=1.0):
        parts = list(
            map(self.__class__, self.chromatogram.split_sparse(delta_rt)))
        for hit in self.tandem_solutions:
            nearest = None
            nearest_time = None
            time = hit.scan_time
            for part in parts:
                time_err = min(abs(part.start_time - time),
                               abs(part.end_time - time))
                if time_err < nearest_time:
                    nearest = part
                    nearest_time = time_err
                if part.spans_time_point(time):
                    part.add_solution(hit)
                    break
            else:
                nearest.add_solution(hit)
        return parts

    def add_solution(self, item: SpectrumSolutionSet):
        case_mass = item.precursor_information.neutral_mass
        if abs(case_mass - self.chromatogram.neutral_mass) > 100:
            log_handle.log(
                "Warning, mis-assigned spectrum match to chromatogram %r, %r" % (self, item))
        self.tandem_solutions.append(item)

    def add_displaced_solution(self, item: SpectrumSolutionSet):
        self.add_solution(item)

    def clone(self) -> 'TandemAnnotatedChromatogram':
        new = super(TandemAnnotatedChromatogram, self).clone()
        new.tandem_solutions = list(self.tandem_solutions)
        new.time_displaced_assignments = list(self.time_displaced_assignments)
        new.best_msms_score = self.best_msms_score
        return new

    def merge(self, other: 'TandemAnnotatedChromatogram', node_type: MassShiftBase = Unmodified) -> 'TandemAnnotatedChromatogram':
        if isinstance(other, TandemAnnotatedChromatogram) or hasattr(other, 'tandem_solutions'):
            new = self.__class__(self.chromatogram.merge(
                other.chromatogram, node_type=node_type))
            new.tandem_solutions = self.tandem_solutions + other.tandem_solutions
            new.time_displaced_assignments = self.time_displaced_assignments + \
                other.time_displaced_assignments
        else:
            new = self.chromatogram.merge(other, node_type=node_type)
        return new

    def merge_in_place(self, other: 'TandemAnnotatedChromatogram', node_type: MassShiftBase = Unmodified):
        if isinstance(other, TandemAnnotatedChromatogram) or hasattr(other, 'tandem_solutions'):
            new = self.chromatogram.merge(
                other.chromatogram, node_type=node_type)
            self.tandem_solutions = self.tandem_solutions + other.tandem_solutions
            self.time_displaced_assignments = self.time_displaced_assignments + \
                other.time_displaced_assignments
        else:
            new = self.chromatogram.merge(other, node_type=node_type)
        self.chromatogram = new

    def assign_entity(self, solution_entry, entity_chromatogram_type=GlycopeptideChromatogram):
        entity_chroma = entity_chromatogram_type(
            None,
            self.chromatogram.nodes, self.chromatogram.mass_shifts,
            self.chromatogram.used_as_mass_shift)
        entity_chroma.entity = solution_entry.solution
        if solution_entry.match.mass_shift != Unmodified:
            identified_shift = solution_entry.match.mass_shift
            for node in entity_chroma.nodes.unspool():
                if node.node_type == Unmodified:
                    node.node_type = identified_shift
                else:
                    node.node_type = (node.node_type + identified_shift)
            entity_chroma.invalidate()
        self.chromatogram = entity_chroma
        self.best_msms_score = solution_entry.best_score


class TandemSolutionsWithoutChromatogram(SpectrumMatchSolutionCollectionBase):
    # For mimicking EntityChromatograms
    entity: Any
    composition: Any

    mass_shift: MassShiftBase

    def __init__(self, entity, tandem_solutions):
        self.entity = entity
        self.composition = entity
        self.tandem_solutions = tandem_solutions
        self.mass_shift = None
        self.best_msms_score = None
        self.update_mass_shift_and_score()

    def update_mass_shift_and_score(self):
        match = self.best_match_for(self.structure)
        if match is not None:
            self.mass_shift = match.mass_shift
            self.best_msms_score = match.score

    def assign_entity(self, solution_entry: SolutionEntry):
        entity = solution_entry.solution
        mass_shift = solution_entry.match.mass_shift
        self.structure = entity
        self.mass_shift = mass_shift
        self.best_msms_score = solution_entry.best_score

    @property
    def mass_shifts(self):
        return [self.mass_shift]

    @property
    def structure(self):
        return self.entity

    @structure.setter
    def structure(self, value):
        self.entity = value
        self.composition = value

    def __repr__(self):
        template = "{self.__class__.__name__}({self.entity!r}, {self.tandem_solutions}, {self.mass_shift})"
        return template.format(self=self)

    @classmethod
    def aggregate(cls, solutions: List[SpectrumSolutionSet]) -> List['TandemSolutionsWithoutChromatogram']:
        collect = DefaultDict(list)
        for solution in solutions:
            best_match = solution.best_solution()
            collect[best_match.target.id].append(solution)
        out = []
        for group in collect.values():
            solution = group[0]
            best_match = solution.best_solution()
            structure = best_match.target
            out.append(cls(structure, group))
        return out


class ScanTimeBundle(object):
    solution: SpectrumSolutionSet
    scan_time: float

    def __init__(self, solution, scan_time):
        self.solution = solution
        self.scan_time = scan_time

    @property
    def score(self):
        try:
            return self.solution.score
        except AttributeError:
            return None

    def __hash__(self):
        return hash((self.solution, self.scan_time))

    def __eq__(self, other):
        return self.solution == other.solution and self.scan_time == other.scan_time

    def __repr__(self):
        return "ScanTimeBundle(%s, %0.4f, %0.4f)" % (
            self.solution.scan.id, self.score, self.scan_time)



from typing import TYPE_CHECKING, List, Optional

import numpy as np

from matplotlib import pyplot as plt, font_manager, axes
from matplotlib import patheffects as path_effects


from six import PY3

from ms_peak_picker.utils import draw_peaklist

from ms_deisotope.data_source.scan.base import ChargeNotProvided
from ms_deisotope.data_source import ProcessedScan

from .sequence_fragment_logo import IonAnnotationGlyphBase, glycopeptide_match_logo, GlycopeptideStubFragmentGlyph, PeptideFragmentGlyph, OxoniumIonGlyph

if TYPE_CHECKING:
    from glycresoft.tandem.spectrum_match.spectrum_match import SpectrumMatcherBase


default_ion_series_to_color = {
    "y": "red",
    "z": "maroon",
    "b": "blue",
    "c": "navy",
    "B": "blue",
    "Y": "red",
    "oxonium_ion": "green",
    "stub_glycopeptide": "goldenrod",
    "precursor": "orange",
}


monosaccharide_to_symbol = {
    "Hex": "\u25CB",
    "HexNAc": "\u25A1",
    "Fuc": "\u25B3",
    "dHex": "\u25B3",
    "HexN": "\u25E9",
    "NeuAc": "\u25C6",
    "NeuGc": "\u25C7",
}


def format_stub_annotation(frag):
    """
    An amusing diversion with unicode, but impractical for
    narrow spectrum plots.
    """
    stack = []
    base = ["Hex", "HexNAc"]
    for k in sorted(frag.glycosylation, key=lambda x: x.mass(), reverse=True):
        if k not in base:
            base.append(k)
    for k in base:
        v = frag.glycosylation[k]
        if not v:
            continue
        stack.append(f" {monosaccharide_to_symbol[k]}{v}")
    stack.append("Pep")
    return "\n".join(stack)


if PY3:
    font_options = font_manager.FontProperties(family="sans serif")
    font_options.set_math_fontfamily("dejavusans")
else:
    font_options = font_manager.FontProperties(family="sans serif")


unicode_superscript = {
    "0": "\u2070",
    "1": "\u00B9",
    "2": "\u00B2",
    "3": "\u00B3",
    "4": "\u2074",
    "5": "\u2075",
    "6": "\u2076",
    "7": "\u2077",
    "8": "\u2078",
    "9": "\u2079",
    "+": "\u207A",
    "-": "\u207B",
}


def encode_superscript(number: int) -> str:
    return "".join([unicode_superscript[c] for c in str(number)])


class SpectrumMatchAnnotator(object):
    usemathtext: bool
    spectrum_match: "SpectrumMatcherBase"
    ax: axes.Axes
    clip_labels: bool
    upper: float
    upshift: float
    peak_labels: List[plt.Text]
    intensity_scale: float
    use_glyphs: bool

    def __init__(
        self,
        spectrum_match: "SpectrumMatcherBase",
        ax: Optional[axes.Axes] = None,
        clip_labels: bool = True,
        usemathtext: bool = False,
        normalize: bool = False,
        use_glyphs=False,
    ):
        if ax is None:
            _, ax = plt.subplots(1)
        self.spectrum_match = spectrum_match
        self.ax = ax
        self.clip_labels = clip_labels
        self.upper = max(spectrum_match.spectrum, key=lambda x: x.intensity).intensity * 1.35
        self.peak_labels = []
        self.upshift = 10
        self.sequence_logo = None
        self.usemathtext = usemathtext
        self.normalize = normalize
        self.intensity_scale = 1.0
        self.use_glyphs = use_glyphs
        if self.use_glyphs:
            self.upper = max(spectrum_match.spectrum, key=lambda x: x.intensity).intensity * 1.0
        if self.normalize:
            self.intensity_scale = max(spectrum_match.spectrum, key=lambda x: x.intensity).intensity
        self.compute_scale()

    def compute_scale(self):
        peaks = self.spectrum_match.deconvoluted_peak_set
        mzs = [p.mz for p in peaks]
        if not mzs:
            self.xscale = 1.0
        else:
            self.xscale = max(mzs) - min(mzs)
        self.yscale = self.upper

    def draw_all_peaks(self, color="grey", **kwargs):
        draw_peaklist(self.spectrum_match.deconvoluted_peak_set, alpha=0.3, color=color, ax=self.ax, **kwargs)
        try:
            draw_peaklist(self.spectrum_match._sanitized_spectrum, color=color, ax=self.ax, alpha=0.5, **kwargs)
        except AttributeError:
            pass

    def add_summary_labels(self, x=0.95, y=0.9):
        prec_purity = self.spectrum_match.scan.annotations.get("precursor purity")
        if prec_purity is not None:
            prec_purity = "%0.2f" % prec_purity
        else:
            prec_purity = "-"

        prec_z = self.spectrum_match.precursor_information.charge
        if prec_z is None or prec_z == ChargeNotProvided:
            prec_z = "-"
        else:
            prec_z = str(prec_z)

        self.ax.text(
            x,
            y,
            "Isolation Purity: %s\nPrec. Z: %s" % (prec_purity, prec_z),
            transform=self.ax.transAxes,
            ha="right",
            color="darkslategrey",
        )

    def label_peak(self, fragment, peak, fontsize=12, rotation=90, **kw):
        label = f"{fragment.name}"
        if peak.charge > 1:
            if self.usemathtext:
                label += f"$^{peak.charge}$"
            else:
                label += encode_superscript(peak.charge)
        y = peak.intensity
        upshift = self.upshift
        if self.use_glyphs:
            y = y + upshift
        else:
            y = min(y + upshift, self.upper * 0.9)

        kw.setdefault("clip_on", self.clip_labels)
        clip_on = kw["clip_on"]

        if self.use_glyphs and fragment.series == "stub_glycopeptide":
            art = GlycopeptideStubFragmentGlyph(
                peak.mz,
                y if not self.normalize else y / self.intensity_scale,
                self.ax,
                fragment,
                self.xscale,
                self.yscale,
                peak.charge,
                size=fontsize * 2,
            )
            art.render()
            self.peak_labels.append(art)
            return art
        elif self.use_glyphs and fragment.series in ("b", "y", "c", "z"):
            art = PeptideFragmentGlyph(
                peak.mz,
                y if not self.normalize else y / self.intensity_scale,
                self.ax,
                fragment,
                self.xscale,
                self.yscale,
                peak.charge,
                size=fontsize * 2,
            )
            art.render()
            self.peak_labels.append(art)
            return art
        elif self.use_glyphs and fragment.series == "oxonium_ion":
            art = OxoniumIonGlyph(
                peak.mz,
                y if not self.normalize else y / self.intensity_scale,
                self.ax,
                fragment,
                self.xscale,
                self.yscale,
                peak.charge,
                size=fontsize * 2
            )
            art.render()
            self.peak_labels.append(art)
            return art
        else:
            text = self.ax.text(
                peak.mz,
                y if not self.normalize else y / self.intensity_scale,
                label,
                rotation=rotation,
                va="bottom",
                ha="center",
                fontsize=fontsize,
                fontproperties=font_options,
                clip_on=clip_on,
                parse_math=self.usemathtext,
            )
            self.peak_labels.append(text)
            return text

    def format_axes(self):
        draw_peaklist([], self.ax, pretty=True)
        bboxes = list(map(lambda x: x.bbox().ymax, filter(lambda x: isinstance(x, IonAnnotationGlyphBase), self.peak_labels)))
        if bboxes:
            upper = max(max(bboxes), self.upper)
        else:
            upper = self.upper

        self.ax.set_ylim(0, upper)

    def draw_matched_peaks(self, color="red", alpha=0.8, fontsize=12, ion_series_to_color=None, **kwargs):
        if ion_series_to_color is None:
            ion_series_to_color = {}
        try:
            solution_map = self.spectrum_match.solution_map
        except AttributeError:
            return
        for peak, fragment in solution_map:
            try:
                peak_color = ion_series_to_color.get(fragment.series, color)
            except AttributeError:
                peak_color = ion_series_to_color.get(fragment.kind, color)
            if self.normalize:
                peak = peak.clone()
                peak.intensity / self.intensity_scale
                # peak = (peak.mz, peak.intensity / self.intensity_scale)
            draw_peaklist([peak], alpha=alpha, ax=self.ax, color=peak_color)
            self.label_peak(fragment, peak, fontsize=fontsize, **kwargs)

    def draw_spectrum_graph(self, color="red", alpha=0.8, fontsize=12, **kwargs):
        try:
            graph = self.spectrum_match.spectrum_graph
        except AttributeError:
            return

        paths = graph.longest_paths(limit=100)

        for path in paths:
            for edge in path:
                self.draw_peak_pair((edge.start, edge.end), color, alpha, fontsize, label=edge.annotation, **kwargs)

    def draw_peak_pair(self, pair, color="red", alpha=0.8, fontsize=12, label=None, rotation=45, **kwargs):
        p1, p2 = pair
        self.ax.plot((p1.mz, p2.mz), (p1.intensity, p2.intensity), color=color, alpha=alpha, **kwargs)
        kwargs.setdefault("clip_on", self.clip_labels)
        clip_on = kwargs["clip_on"]
        draw_peaklist(pair, ax=self.ax, alpha=0.4, color="orange")
        if label:
            # pylint: disable=assignment-from-no-return
            midx = (p1.mz + p2.mz) / 2
            # interpolate the midpoint's height
            midy = (p1.intensity * (p2.mz - midx) + p2.intensity * (midx - p1.mz)) / (p2.mz - p1.mz)

            # find the angle of the line connecting the two peaks
            xlo = min(p1.mz, p2.mz)
            xhi = max(p1.mz, p2.mz)
            adj = xhi - xlo
            ylo = min(p1.intensity, p2.intensity)
            yhi = max(p1.intensity, p2.intensity)
            opp = yhi - ylo
            hypot = np.hypot(adj, opp)
            rotation = np.arccos(adj / hypot)

            if isinstance(label, (list, tuple)):
                label = "-".join(map(str, label))
            else:
                label = str(label)
            text = self.ax.text(
                midx, midy, label, fontsize=fontsize, ha="center", va="bottom", rotation=rotation, clip_on=clip_on
            )
            text.set_path_effects(
                [
                    path_effects.Stroke(linewidth=0.5, foreground="white"),
                    path_effects.Normal(),
                ]
            )

    def draw(self, **kwargs):
        fontsize = kwargs.pop("fontsize", 9)
        rotation = kwargs.pop("rotation", 90)
        clip_labels = kwargs.pop("clip_labels", self.clip_labels)
        self.clip_labels = clip_labels
        ion_series_to_color = kwargs.pop("ion_series_to_color", default_ion_series_to_color)
        self.draw_all_peaks(**kwargs)
        self.draw_matched_peaks(fontsize=fontsize, ion_series_to_color=ion_series_to_color, rotation=rotation, **kwargs)
        self.draw_spectrum_graph(fontsize=fontsize, rotation=rotation / 2)
        self.format_axes()
        return self

    def add_logo_plot(self, xrel=0.15, yrel=0.8, width=0.67, height=0.13, **kwargs):
        figure = self.ax.figure
        iax = figure.add_axes([xrel, yrel, width, height])
        logo = glycopeptide_match_logo(self.spectrum_match, ax=iax, **kwargs)
        self.sequence_logo = logo
        return logo

    def _draw_mass_accuracy_plot(self, ax, error_tolerance=2e-5, **kwargs):
        ion_series_to_color = kwargs.pop("ion_series_to_color", default_ion_series_to_color)
        match = self.spectrum_match
        ax.scatter(
            *zip(*[(pp.peak.mz, pp.mass_accuracy()) for pp in match.solution_map]),
            alpha=0.5,
            edgecolor="black",
            color=[ion_series_to_color[pp.fragment.series] for pp in match.solution_map],
        )
        limits = error_tolerance
        ax.set_ylim(-limits, limits)
        xlim = 0, max(match.deconvoluted_peak_set, key=lambda x: x.mz).mz + 100
        ax.hlines(0, *xlim, linestyle="--", color="black", lw=0.75)
        ax.hlines(limits / 2, *xlim, linestyle="--", lw=0.5, color="black")
        ax.hlines(-limits / 2, *xlim, linestyle="--", lw=0.5, color="black")
        ax.set_xlim(*xlim)
        labels = ax.get_yticks()
        labels = ["%0.2g" % (label * 1e6) for label in labels]
        ax.set_yticklabels(labels)
        ax.set_xlabel("m/z", fontsize=10)
        ax.set_ylabel("Mass Accuracy (PPM)", fontsize=10)
        return ax

    def __repr__(self):
        return f"{self.__class__.__name__}({self.spectrum_match})"


class TidySpectrumMatchAnnotator(SpectrumMatchAnnotator):
    def label_peak(self, fragment, peak, fontsize=12, rotation=90, **kw):
        min_intensity = 0.02 * (self.upper / 1.35)
        if fragment.series == "oxonium_ion":
            if peak.intensity < min_intensity:
                return
        super(TidySpectrumMatchAnnotator, self).label_peak(fragment, peak, fontsize, rotation, **kw)


def normalize_scan(scan: ProcessedScan, factor=None):
    scan = scan.copy()
    scan.annotations.pop("peak_label_map", None)
    scan.deconvoluted_peak_set = scan.deconvoluted_peak_set.__class__(p.clone() for p in scan.deconvoluted_peak_set)
    bp = scan.base_peak()
    if factor is None:
        factor = bp.intensity / 1000
    for peak in scan:
        peak.intensity /= factor
    return scan


class MirrorSpectrumAnnotatorFacet(TidySpectrumMatchAnnotator):
    def draw_all_peaks(self, color="black", alpha=0.5, mirror=False, **kwargs):
        draw_peaklist(
            self.spectrum_match.deconvoluted_peak_set
            if not mirror
            else [[p.mz, -p.intensity] for p in self.spectrum_match.deconvoluted_peak_set],
            alpha=0.3,
            color="grey",
            ax=self.ax,
            lw=0.75,
            **kwargs,
        )

    def draw_matched_peaks(self, color="red", alpha=0.8, fontsize=12, ion_series_to_color=None, mirror=False, **kwargs):
        if ion_series_to_color is None:
            ion_series_to_color = {}
        try:
            solution_map = self.spectrum_match.solution_map
        except AttributeError:
            return
        for peak, fragment in solution_map:
            try:
                peak_color = ion_series_to_color.get(fragment.series, color)
            except AttributeError:
                peak_color = ion_series_to_color.get(fragment.kind, color)
            draw_peaklist(
                [peak] if not mirror else [(peak.mz, -peak.intensity)], alpha=alpha, ax=self.ax, color=peak_color
            )
            self.label_peak(fragment, peak, fontsize=fontsize, mirror=mirror, **kwargs)

    def base_peak_factor(self):
        return self.spectrum_match.scan.base_peak().intensity / 100

    def label_peak(self, fragment, peak, fontsize=12, rotation=90, mirror=False, **kw):
        label = "%s" % fragment.name
        if fragment.series == "oxonium_ion":
            return ""
        if peak.charge > 1:
            if self.usemathtext:
                label += f"$^{peak.charge}$"
            else:
                label += encode_superscript(peak.charge)
        y = peak.intensity
        upshift = 2
        sign = 1 if y > 0 else -1
        y = min(y + sign * upshift, self.upper * 0.9)

        if mirror:
            y = -y

        kw.setdefault("clip_on", self.clip_labels)
        clip_on = kw["clip_on"]

        text = self.ax.text(
            peak.mz,
            y,
            label,
            rotation=rotation,
            va="bottom" if not mirror else "top",
            ha="center",
            fontsize=fontsize,
            fontproperties=font_options,
            clip_on=clip_on,
        )
        self.peak_labels.append(text)
        return text

    def draw(self, mirror=False, **kwargs):
        fontsize = kwargs.pop("fontsize", 9)
        rotation = kwargs.pop("rotation", 90)
        clip_labels = kwargs.pop("clip_labels", self.clip_labels)
        self.clip_labels = clip_labels
        ion_series_to_color = kwargs.pop("ion_series_to_color", default_ion_series_to_color)
        self.draw_all_peaks(mirror=mirror, **kwargs)
        self.draw_matched_peaks(
            fontsize=fontsize, ion_series_to_color=ion_series_to_color, rotation=rotation, mirror=mirror, **kwargs
        )
        self.draw_spectrum_graph(fontsize=fontsize, rotation=rotation / 2)
        self.format_axes()
        return self


def mirror_spectra(psm_a, psm_b):
    art = MirrorSpectrumAnnotatorFacet(psm_a)
    art.draw()
    reflect = MirrorSpectrumAnnotatorFacet(psm_b, ax=art.ax)
    reflect.draw(mirror=True)
    reflect.ax.set_ylim(-1500, 1200)
    ax = art.ax
    ax.set_yticks(np.arange(-1000, 1000 + 250, 250))
    ax.set_yticklabels(map(lambda x: str(x) + "%", list(range(100, -25, -25)) + list(range(25, 125, 25))))
    return art.ax, art, reflect



from array import ArrayType as array
from concurrent.futures import ThreadPoolExecutor

from typing import Dict, Optional, Protocol, Union, List

import numpy as np
from numpy.typing import ArrayLike

from scipy import stats
from scipy.special import logsumexp

try:
    from matplotlib import pyplot as plt
except ImportError:
    pass


class KMeans(object):
    k: int
    means: np.ndarray

    def __init__(self, k, means=None):
        self.k = k
        self.means = np.array(means) if means is not None else None

    @classmethod
    def fit(cls, X: np.ndarray, k: int, initial_mus: Optional[np.ndarray]=None):
        if initial_mus is not None:
            mus = initial_mus
        else:
            mus = np.sort(np.random.choice(X, k))
        inst = cls(k, mus)
        inst.estimate(X)
        return inst

    @classmethod
    def from_json(cls, state: Dict) -> 'KMeans':
        return cls(state['k'], state['means'])

    def to_json(self) -> Dict:
        return {
            "k": int(self.k),
            "means": [float(m) for m in self.means]
        }

    def estimate(self, X: np.ndarray, maxiter: int=1000, tol: float=1e-6):
        for i in range(maxiter):
            distances = []
            for k in range(self.k):
                diff = (X - self.means[k])
                dist = np.sqrt((diff * diff))
                distances.append(dist)
            distances = np.vstack(distances).T
            cluster_assignments = np.argmin(distances, axis=1)
            new_means = []
            for k in range(self.k):
                new_means.append(
                    np.mean(X[cluster_assignments == k]))
            new_means = np.array(new_means)
            new_means[np.isnan(new_means)] = 0.0
            diff = (self.means - new_means)
            dist = np.sqrt((diff * diff).sum()) / self.k
            self.means = new_means
            # Always sort the means so that ordering is consistent
            self.means.sort()
            if dist < tol:
                break
        else:
            pass

    def score(self, x: Union[float, np.ndarray]) -> np.ndarray:
        x = np.asanyarray(x)
        if x.ndim < 2:
            x = x.reshape((-1, 1))
        delta = np.abs(self.means - x)
        score = 1 - delta
        norm = score.sum(axis=1)[:, None] + 1e-13
        score /= norm
        return score

    def predict(self, x: Union[float, np.ndarray]) -> np.ndarray:
        scores = self.score(x)
        return np.argmax(scores, axis=1)


class MixtureBase(object):
    n_components: int

    def __init__(self, n_components):
        self.n_components

    def to_json(self) -> Dict:
        return {}

    @classmethod
    def from_json(cls, state) -> 'MixtureBase':
        raise NotImplementedError()

    def loglikelihood(self, X) -> float:
        out = logsumexp(self.logpdf(X), axis=1).sum()
        return out

    def bic(self, X) -> float:
        '''Calculate the Bayesian Information Criterion
        for selecting the most parsimonious number of components.
        '''
        return np.log(X.size) * (self.n_components * 3 - 1) - (2 * (self.loglikelihood(X)))

    def logpdf(self, X: np.ndarray, weighted: float=True) -> np.ndarray:
        out = np.array(
            [self._logpdf(X, k)
             for k in range(self.n_components)]).T
        if weighted:
            out += np.log(self.weights)
        return out

    def pdf(self, X: np.ndarray, weighted: float = True) -> np.ndarray:
        return np.exp(self.logpdf(X, weighted=weighted))

    def score(self, X: np.ndarray) -> np.ndarray:
        return self.pdf(X).sum(axis=1)

    def responsibility(self, X: np.ndarray) -> np.ndarray:
        '''Also called the posterior probability, as these are the
        probabilities associating each element of X with each component
        '''
        acc = np.zeros((X.shape[0], self.n_components))
        for k in range(self.n_components):
            acc[:, k] = np.log(self.weights[k]) + self._logpdf(X, k)
        total = logsumexp(acc, axis=1)[:, None]
        # compute the ratio of the density to the total in log-space, then
        # exponentiate to return to linear space
        out = np.exp(acc - total)
        return out


class GaussianMixture(MixtureBase):
    mus: np.ndarray
    sigmas: np.ndarray
    weights: np.ndarray

    def __init__(self, mus, sigmas, weights):
        self.mus = np.array(mus)
        self.sigmas = np.array(sigmas)
        self.weights = np.array(weights)
        self.n_components = len(weights)

    @classmethod
    def from_json(cls, state: Dict) -> 'GaussianMixture':
        return cls(state['mus'], state['sigmas'], state['weights'])

    def to_json(self) -> Dict:
        return {
            "mus": [float(m) for m in self.mus],
            "sigmas": [float(m) for m in self.sigmas],
            "weights": [float(m) for m in self.weights]
        }

    def __repr__(self):
        template = "{self.__class__.__name__}({self.mus}, {self.sigmas}, {self.weights})"
        return template.format(self=self)

    def _logpdf(self, X: np.ndarray, k: int) -> np.ndarray:
        '''Computes the log-space density for `X` using the `k`th
        component of the mixture
        '''
        return stats.norm.logpdf(X, self.mus[k], self.sigmas[k])

    @classmethod
    def fit(cls, X: np.ndarray, n_components: int, maxiter: int=1000, tol: float=1e-5, deterministic: bool=True) -> 'GaussianMixture':
        if not deterministic:
            mus = KMeans.fit(X, n_components).means
        else:
            mus = (np.max(X) / (n_components + 1)) * np.arange(1, n_components + 1)
        assert not np.any(np.isnan(mus))
        sigmas = np.var(X) * np.ones_like(mus)
        weights = np.ones_like(mus) / n_components
        inst = cls(mus, sigmas, weights)
        inst.estimate(X, maxiter=maxiter, tol=tol)
        return inst

    def _update_params_for(self, X: np.ndarray, k: int, responsibility: np.ndarray, new_mus: np.ndarray,
                           new_sigmas: np.ndarray, new_weights: np.ndarray):
        # The expressions for each partial derivative may be useful for understanding
        # portions of this block.
        # See http://www.notenoughthoughts.net/posts/normal-log-likelihood-gradient.html
        g = responsibility[:, k]
        N_k = g.sum()
        # Begin specialization for Gaussian distributions
        diff = X - self.mus[k]
        mu_k = g.dot(X) / N_k
        new_mus[k] = mu_k
        sigma_k = (g * diff).dot(diff.T) / N_k + 1e-6
        new_sigmas[k] = np.sqrt(sigma_k)
        new_weights[k] = N_k

    def estimate(self, X: np.ndarray, maxiter: int=1000, tol: float=1e-5):
        for i in range(maxiter):
            # E-step
            responsibility = self.responsibility(X)

            # M-step
            new_mus = np.zeros_like(self.mus)
            new_sigmas = np.zeros_like(self.sigmas)
            prev_loglikelihood = self.loglikelihood(X)
            new_weights = np.zeros_like(self.weights)

            for k in range(self.n_components):
                self._update_params_for(X, k, responsibility, new_mus, new_sigmas, new_weights)

            new_weights /= new_weights.sum()
            self.mus = new_mus
            self.sigmas = new_sigmas
            self.weights = new_weights
            new_loglikelihood = self.loglikelihood(X)
            delta_fit = (prev_loglikelihood - new_loglikelihood) / new_loglikelihood
            if abs(delta_fit) < tol:
                break
        else:
            pass

    @property
    def domain(self):
        return [
            self.mus.min() - self.sigmas.max() * 4,
            self.mus.max() + self.sigmas.max() * 4
         ]

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(1)
        X = np.arange(*self.domain, step=0.01)
        Y = np.exp(self.logpdf(X, True))
        ax.plot(X, np.sum(Y, axis=1), **kwargs)
        return ax


a = 0.0
b = float('inf')
phi_a = stats.norm.pdf(a)
phi_b = stats.norm.pdf(b)
Z = stats.norm.cdf(b) - stats.norm.cdf(a)


def _truncnorm_mean(X):
    mu = X.mean()
    sigma = np.std(X)
    return mu + (phi_a / Z) * sigma


def _truncnorm_std(X):
    sigma2 = np.var(X)
    return np.sqrt(sigma2 * (1 + a * phi_a / Z - (phi_a / Z) ** 2))


def truncnorm_pdf(x, mu, sigma):
    scalar = np.isscalar(x)
    if scalar:
        x = np.array([x])
    mask = (a <= x) & (x <= b)
    out = np.zeros_like(x)

    numerator = stats.norm.pdf((x[mask] - mu) / sigma)
    denominator = stats.norm.cdf(
        (b - mu) / sigma) - stats.norm.cdf((a - mu) / sigma)
    out[mask] = 1 / sigma * numerator / denominator
    if scalar:
        out = out[0]
    return out


def truncnorm_logpdf(x, mu, sigma):
    return np.log(truncnorm_pdf(x, mu, sigma))


class _TruncatedNormalMixin(object):

    def _logpdf(self, X: np.ndarray, k: int) -> np.ndarray:
        '''Computes the log-space density for `X` using the `k`th
        component of the mixture
        '''
        result = truncnorm_logpdf(
            X, self.mus[k], self.sigmas[k])
        return result

    def _update_params_for(self, X, k, responsibility, new_mus, new_sigmas, new_weights):
        # The expressions for each partial derivative may be useful for understanding
        # portions of this block.
        # See http://www.notenoughthoughts.net/posts/normal-log-likelihood-gradient.html
        g = responsibility[:, k]
        N_k = g.sum()

        # Begin specialization for the truncated Gaussian distributions
        diff = X - self.mus[k]

        unconstrained_mu_k = g.dot(X) / N_k
        unconstrained_sigma_k = np.sqrt((g * diff).dot(diff.T) / N_k + 1e-6)
        mu_k = unconstrained_mu_k + (phi_a / Z) * unconstrained_sigma_k
        sigma_k = np.sqrt(unconstrained_sigma_k ** 2 * (1 + a * phi_a / Z - (phi_a / Z) ** 2))

        new_mus[k] = mu_k
        new_sigmas[k] = np.sqrt(sigma_k)
        new_weights[k] = N_k


class TruncatedGaussianMixture(_TruncatedNormalMixin, GaussianMixture):
    @classmethod
    def fit(cls, X, n_components, maxiter=1000, tol=1e-5, deterministic=True) -> 'TruncatedGaussianMixture':
        if not deterministic:
            mus = KMeans.fit(X, n_components).means
        else:
            mus = (np.max(X) / (n_components + 1)) * \
                np.arange(0, n_components + 1)
        mus[0] = 0.0
        assert not np.any(np.isnan(mus))
        sigmas = np.var(X) * np.ones_like(mus)
        weights = np.ones_like(mus) / n_components
        inst = cls(mus, sigmas, weights)
        inst.estimate(X, maxiter=maxiter, tol=tol)
        return inst


class GammaMixtureBase(MixtureBase):
    shapes: np.ndarray
    scales: np.ndarray
    weights: np.ndarray

    def __init__(self, shapes, scales, weights):
        self.shapes = np.array(shapes)
        self.scales = np.array(scales)
        self.weights = np.array(weights)
        self.n_components = len(weights)

    @classmethod
    def from_json(cls, state: Dict) -> 'GammaMixtureBase':
        return cls(state['shapes'], state['scales'], state['weights'])

    def to_json(self) -> Dict:
        return {
            "shapes": [float(m) for m in self.shapes],
            "scales": [float(m) for m in self.scales],
            "weights": [float(m) for m in self.weights]
        }

    def __repr__(self):
        template = "{self.__class__.__name__}({self.shapes}, {self.scales}, {self.weights})"
        return template.format(self=self)

    def _logpdf(self, X: np.ndarray, k: int) -> np.ndarray:
        """
        Computes the log-space density for `X` using the `k`th
        component of the mixture
        """
        return stats.gamma.logpdf(X, a=self.shapes[k], scale=self.scales[k])

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(1)
        X = np.arange(*self.domain, step=0.01)
        Y = np.exp(self.logpdf(X, True))
        ax.plot(X, np.sum(Y, axis=1), **kwargs)
        return ax

    @property
    def domain(self):
        return [
            1e-6,
            100.0
        ]

    @classmethod
    def fit(cls, X: np.ndarray, n_components: int, maxiter: int=100, tol: float=1e-5, deterministic: bool=True) -> 'GammaMixtureBase':
        shapes, scales, weights = cls.initial_parameters(X, n_components, deterministic=deterministic)
        inst = cls(shapes, scales, weights)
        inst.estimate(X, maxiter=maxiter, tol=tol)
        return inst


class IterativeGammaMixture(GammaMixtureBase):
    '''An iterative approximation of a mixture Gamma distributions
    based on the Gaussian distribution. May not converge to the optimal
    solution, and if so, it converges slowly.

    Derived from pGlyco's FDR estimation method
    '''
    @staticmethod
    def initial_parameters(X, n_components, deterministic=True):
        mu = np.median(X) / (n_components + 1) * np.arange(1, n_components + 1)
        sigma = np.ones(n_components) * np.var(X)
        shapes = mu ** 2 / sigma
        scales = sigma / mu
        weights = np.ones(n_components)
        weights /= weights.sum()
        return shapes, scales, weights

    def estimate(self, X: np.ndarray, maxiter=100, tol=1e-5):
        prev_loglikelihood = self.loglikelihood(X)
        for i in range(maxiter):
            # E-Step
            responsibility = self.responsibility(X)

            # M-Step
            new_weights = responsibility.sum(axis=0) / responsibility.sum()
            mu = responsibility.T.dot(X) / responsibility.T.sum(axis=1) + 1e-6
            sigma = np.array(
                [responsibility[:, i].dot((X - mu[i]) ** 2 / np.sum(responsibility[:, i]))
                 for i in range(self.n_components)]) + 1e-6
            new_shapes = mu ** 2 / sigma
            new_scales = sigma / mu
            self.shapes = new_shapes
            self.scales = new_scales
            self.weights = new_weights

            new_loglikelihood = self.loglikelihood(X)
            delta_fit = (prev_loglikelihood - new_loglikelihood) / new_loglikelihood
            if abs(delta_fit) < tol:
                break
        else:
            pass


GammaMixture = IterativeGammaMixture


class GaussianMixtureWithPriorComponent(GaussianMixture):
    prior: MixtureBase

    def __init__(self, mus, sigmas, prior, weights):
        self.mus = np.array(mus)
        self.sigmas = np.array(sigmas)
        self.prior = prior
        self.weights = np.array(weights)
        self.n_components = len(weights)

    def to_json(self) -> Dict:
        state = super().to_json()
        state['prior'] = self.prior.to_json()
        state['prior_type'] = self.prior.__class__.__name__
        return state

    @classmethod
    def from_json(cls, state: Dict) -> 'GaussianMixtureWithPriorComponent':
        # TODO: Need to derive the type object of the prior from its name
        # prior_type_name = state['prior_type']
        prior = GammaMixtureBase.from_json(state['prior'])
        return cls(state['mus'], state['sigmas'], prior, state['weights'])

    def _logpdf(self, X, k):
        if k == self.n_components - 1:
            return np.log(np.exp(self.prior.logpdf(X, weighted=False)).dot(self.prior.weights))
        else:
            return super(GaussianMixtureWithPriorComponent, self)._logpdf(X, k)

    @classmethod
    def fit(cls, X: np.ndarray, n_components: int, prior: MixtureBase, maxiter=1000, tol=1e-5, deterministic=True) -> 'GaussianMixtureWithPriorComponent':
        if not deterministic:
            mus = KMeans.fit(X, n_components).means
        else:
            mus = (np.max(X) / (n_components + 1)) * np.arange(1, n_components + 1)
        assert not np.any(np.isnan(mus))
        sigmas = np.var(X) * np.ones_like(mus)
        weights = np.ones(n_components + 1) / (n_components + 1)
        inst = cls(mus, sigmas, prior, weights)
        inst.estimate(X, maxiter=maxiter, tol=tol)
        return inst

    def estimate(self, X: np.ndarray, maxiter=1000, tol=1e-5):
        for i in range(maxiter):
            # E-step
            responsibility = self.responsibility(X)

            # M-step
            new_mus = np.zeros_like(self.mus)
            new_sigmas = np.zeros_like(self.sigmas)
            prev_loglikelihood = self.loglikelihood(X)
            new_weights = np.zeros_like(self.weights)
            for k in range(self.n_components - 1):
                self._update_params_for(
                    X, k, responsibility, new_mus, new_sigmas, new_weights)

            new_weights = responsibility.sum(axis=0) / responsibility.sum()
            self.mus = new_mus
            self.sigmas = new_sigmas
            self.weights = new_weights
            new_loglikelihood = self.loglikelihood(X)
            delta_fit = (prev_loglikelihood - new_loglikelihood) / new_loglikelihood
            if abs(delta_fit) < tol:
                break
        else:
            pass

    def plot(self, ax=None, **kwargs):
        ax = super(GaussianMixtureWithPriorComponent, self).plot(ax=ax, **kwargs)
        X = np.arange(self.prior.domain[0], self.mus.max() + self.sigmas.max() * 4, 0.01)
        Y = self.prior.score(X) * self.weights[-1]
        ax.plot(X, Y, **kwargs)
        return ax


class TruncatedGaussianMixtureWithPriorComponent(_TruncatedNormalMixin, GaussianMixtureWithPriorComponent):

    @classmethod
    def fit(cls, X, n_components, prior, maxiter=1000, tol=1e-5, deterministic=True) -> 'TruncatedGaussianMixtureWithPriorComponent':
        if not deterministic:
            mus = KMeans.fit(X, n_components).means
        else:
            mus = (np.max(X) / (n_components + 1)) * \
                np.arange(1, n_components + 1)
        assert not np.any(np.isnan(mus))
        mus[0] = 0.0
        sigmas = np.var(X) * np.ones_like(mus)
        weights = np.ones(n_components + 1) / (n_components + 1)
        inst = cls(mus, sigmas, prior, weights)
        inst.estimate(X, maxiter=maxiter, tol=tol)
        return inst

    def _logpdf(self, X, k):
        if k == self.n_components - 1:
            return np.log(np.exp(self.prior.logpdf(X, weighted=False)).dot(self.prior.weights))
        else:
            return _TruncatedNormalMixin._logpdf(self, X, k)

    def _update_params_for(self, X, k, responsibility, new_mus, new_sigmas, new_weights):
        # The expressions for each partial derivative may be useful for understanding
        # portions of this block.
        # See http://www.notenoughthoughts.net/posts/normal-log-likelihood-gradient.html
        g = responsibility[:, k]
        N_k = g.sum()
        # Begin specialization for Gaussian distributions
        diff = X - self.mus[k]
        if k == 0:
            mu_k = 0.0
        else:
            mu_k = g.dot(X) / N_k
        new_mus[k] = mu_k
        sigma_k = (g * diff).dot(diff.T) / N_k + 1e-6
        new_sigmas[k] = np.sqrt(sigma_k)
        new_weights[k] = N_k


class PredictorBase(Protocol):
    def predict(self, value: Union[float, ArrayLike]) -> Union[float, ArrayLike]:
        ...


class KDModel(PredictorBase):
    positive_probabilities: array
    negative_probabilities: array
    values: array

    positive_prior: float

    positive_kernel_fit: stats.gaussian_kde
    negative_kernel_fit: stats.gaussian_kde

    def __init__(self,
                 positive_prior: float=None,
                 positive_probabilities: array=None,
                 negative_probabilities: array=None,
                 values: array=None):
        if positive_probabilities is None:
            positive_probabilities = array('d')
        if negative_probabilities is None:
            negative_probabilities = array('d')
        if values is None:
            values = array('d')
        if positive_prior is None:
            positive_prior = 0.5

        self.positive_probabilities = positive_probabilities
        self.negative_probabilities = negative_probabilities
        self.values = values

        self.positive_prior = positive_prior

        self.positive_kernel_fit = None
        self.negative_kernel_fit = None

    def add(self, prob: float, val: float):
        self.positive_probabilities.append(prob)
        self.negative_probabilities.append(1 - prob)
        self.values.append(val)

    def fit(self, maxiter: int=10):
        lastprobsum = self.update()

        i = 0
        while i < maxiter:
            probsum = self.update()
            i += 1
            if abs(probsum - lastprobsum) < 0.001:
                break
            lastprobsum = probsum

    def update_fits(self):
        self.positive_kernel_fit = GaussianKDE(
            self.values, weights=self.positive_probabilities)
        self.negative_kernel_fit = GaussianKDE(
            self.values, weights=self.negative_probabilities)

    def update(self):
        self.update_fits()
        probs = self.predict(self.values)
        count = len(probs)
        probsum = np.sum(probs)
        self.positive_prior = probsum / count
        self.positive_probabilities = probs
        self.negative_probabilities = 1 - probs
        return probsum

    def predict(self, value: float) -> float:
        if self.positive_kernel_fit is None:
            raise TypeError("Cannot predict with a model that has not been fit yet")

        p = self.predict_positive(value)
        n = self.predict_negative(value)

        pos_prior = self.positive_prior
        neg_prior = 1 - pos_prior
        return (p * pos_prior) / ((p * pos_prior) + (n * neg_prior))

    def predict_positive(self, value: float) -> float:
        return self.positive_kernel_fit.pdf(value)

    def predict_negative(self, value: float) -> float:
        return self.negative_kernel_fit.pdf(value)


class GaussianKDE(stats.gaussian_kde):
    '''To control overriding methods, use a derived class'''
    pass


KDE_USE_THREADS = True
try:
    from glycresoft._c.structure.probability import evaluate_gaussian_kde
    GaussianKDE.evaluate = evaluate_gaussian_kde
except ImportError as err:
    print(err)
    KDE_USE_THREADS = False


class MultiKDModel(PredictorBase):
    models: List[KDModel]
    values: List[array]
    positive_prior: float

    use_threads: bool
    thread_pool: Optional[ThreadPoolExecutor]

    def __init__(self,
                 positive_prior: float=None,
                 models: List[KDModel]=None,
                 values: array=None,
                 use_threads: bool=KDE_USE_THREADS):
        if positive_prior is None:
            positive_prior = 0.5
        if models is None:
            models = []
        if values is None:
            values = [array('d') for m in models]
        self.models = models
        self.positive_prior = positive_prior
        self.values = values
        self.use_threads = use_threads
        self.thread_pool = None
        if self.use_threads:
            self.thread_pool = ThreadPoolExecutor()

    def __getstate__(self):
        state = {
            'positive_prior': self.positive_prior,
            'models': self.models,
            'values': self.values,
        }
        return state

    def __setstate__(self, state):
        self.positive_prior = state['positive_prior']
        self.models = state['models']
        self.values = state['values']

    def __reduce__(self):
        return self.__class__, (), self.__getstate__()

    def add_model(self, model: KDModel):
        self.models.append(model)
        self.values.append(model.values)

    def add(self, prob: float, values: List[float]):
        for val, mod, valn in zip(values, self.models, self.values):
            mod.add(prob, val)
            valn.append(val)

    def update_fits(self):
        for model in self.models:
            model.update_fits()

    def close_thread_pool(self):
        if self.use_threads:
            if self.thread_pool is not None:
                self.thread_pool.shutdown()
            self.thread_pool =  None
            self.use_threads = False

    def create_thread_pool(self):
        self.use_threads = True
        self.thread_pool = ThreadPoolExecutor()

    def update(self):
        self.update_fits()
        probs = self.predict(self.values)
        count = len(probs)
        probsum = np.sum(probs)
        self.positive_prior = probsum / count
        for mod in self.models:
            mod.positive_probabilities = probs
            mod.negative_probabilities = 1 - probs
        return probsum

    def predict_positive(self, values: List[float]) -> float:
        acc = 0
        if self.use_threads:
            jobs = []
            for mod, val in zip(self.models, values):
                jobs.append(self.thread_pool.submit(mod.positive_kernel_fit.evaluate, val))

            for job in jobs:
                acc += np.log(job.result())
        else:
            for mod, val in zip(self.models, values):
                acc += np.log(mod.positive_kernel_fit.evaluate(val))
        return np.exp(acc)

    def predict_negative(self, values: List[float]) -> float:
        acc = 0
        if self.use_threads:
            jobs = []
            for mod, val in zip(self.models, values):
                jobs.append(self.thread_pool.submit(
                    mod.negative_kernel_fit.evaluate, val))

            for job in jobs:
                acc += np.log(job.result())
        else:
            for mod, val in zip(self.models, values):
                acc += np.log(mod.negative_kernel_fit.evaluate(val))
        return np.exp(acc)

    def predict(self, values: List[float]) -> float:
        if self.use_threads:
            p_fut = self.thread_pool.submit(self.predict_positive, values)
            n_fut = self.thread_pool.submit(self.predict_negative, values)
            p = p_fut.result()
            n = n_fut.result()
        else:
            p = self.predict_positive(values)
            n = self.predict_negative(values)

        pos_prior = self.positive_prior
        neg_prior = 1 - pos_prior
        return (p * pos_prior) / ((p * pos_prior) + (n * neg_prior))

    def fit(self, maxiter: int=10, convergence: float=0.001):
        lastprobsum = self.update()
        delta = float('inf')
        i = 1
        while i < maxiter:
            probsum = self.update()
            i += 1
            delta = abs(probsum - lastprobsum)
            if delta < convergence:
                break
            lastprobsum = probsum
        return i, delta



from .csv_format import (
    CSVSerializerBase,
    GlycanHypothesisCSVSerializer,
    ImportableGlycanHypothesisCSVSerializer,
    GlycopeptideHypothesisCSVSerializer,
    GlycanLCMSAnalysisCSVSerializer,
    GlycopeptideLCMSMSAnalysisCSVSerializer,
    GlycopeptideSpectrumMatchAnalysisCSVSerializer,
    SimpleChromatogramCSVSerializer,
    SimpleScoredChromatogramCSVSerializer,
    MultiScoreGlycopeptideLCMSMSAnalysisCSVSerializer,
    MultiScoreGlycopeptideSpectrumMatchAnalysisCSVSerializer)

from .xml import (
    MzIdentMLSerializer)

from .report import (
    GlycanChromatogramReportCreator,
    GlycopeptideDatabaseSearchReportCreator)

from .annotate_spectra import SpectrumAnnotatorExport, CSVSpectrumAnnotatorExport


from .text_format import TrainingMGFExporter



from collections import Counter

import numpy as np
from matplotlib import pyplot as plt

from glypy.structure.glycan_composition import HashableGlycanComposition

from .glycan_visual_classification import (
    NGlycanCompositionColorizer,
    NGlycanCompositionOrderer,
    GlycanLabelTransformer)
from .chromatogram_artist import ArtistBase


class EntitySummaryBarChartArtist(ArtistBase):
    bar_width = 0.5
    alpha = 0.5
    y_label = "<SET self.y_label>"
    plot_title = "<SET self.plot_title>"

    def __init__(self, chromatograms, ax=None, colorizer=None, orderer=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        if colorizer is None:
            colorizer = NGlycanCompositionColorizer
        if orderer is None:
            orderer = NGlycanCompositionOrderer
        self.ax = ax
        self.colorizer = colorizer
        self.orderer = orderer
        self.chromatograms = [c for c in chromatograms if c.glycan_composition is not None]

    def sort_items(self):
        return self.orderer.sort(self.chromatograms, key=lambda x: x.glycan_composition)

    def get_heights(self, items, **kwargs):
        raise NotImplementedError()

    def configure_axes(self):
        self.ax.axes.spines['right'].set_visible(False)
        self.ax.axes.spines['top'].set_visible(False)
        self.ax.yaxis.tick_left()
        self.ax.xaxis.set_ticks_position('none')
        self.ax.xaxis.set_ticks_position('none')
        self.ax.set_title(self.plot_title, fontsize=24)
        self.ax.set_ylabel(self.y_label, fontsize=18)

    def __len__(self):
        return len(self.sort_items())

    def prepare_x_args(self):
        items = self.sort_items()
        keys = [c.glycan_composition for c in items]
        include_classes = set(map(self.colorizer.classify, keys))
        xtick_labeler = GlycanLabelTransformer(keys, self.orderer)
        color = list(map(self.colorizer, keys))
        self.indices = indices = np.arange(len(items))

        self.xtick_labeler = xtick_labeler
        self.keys = keys
        self.color = color
        self.include_classes = include_classes
        self.items = items

        return items, keys, include_classes, xtick_labeler, color, indices

    def configure_x_axis(self):
        ax = self.ax
        ax.set_xticks(self.indices + (self.bar_width))
        n = len(self.indices)
        if not n:
            n = 1
        font_size = min(
            max(
                (150. / (n / 2.)),
                3),
            24)

        ax.set_xlabel(self.xtick_labeler.label_key, fontsize=14)
        ax.set_xticklabels(tuple(self.xtick_labeler), rotation=90, ha='center', size=font_size)
        if len(self.indices) == 1:
            lo, hi = ax.get_xlim()
            hi *= 2
            ax.set_xlim(lo, hi)

    def draw(self, logscale=False):
        items, keys, include_classes, xtick_labeler, color, indices = self.prepare_x_args()
        heights = self.get_heights(items, logscale)

        self.bars = self.ax.bar(
            indices + self.bar_width, heights,
            width=self.bar_width, color=color, alpha=self.alpha, lw=0)

        self.configure_x_axis()

        handles = NGlycanCompositionColorizer.make_legend(
            include_classes, alpha=self.alpha)
        if handles:
            self.ax.legend(handles=handles, bbox_to_anchor=(1.20, 1.0))

        self.configure_axes()

        return self


class BundledGlycanComposition(object):
    def __init__(self, glycan_composition, total_signal):
        self.glycan_composition = HashableGlycanComposition.parse(str(glycan_composition))
        self.total_signal = float(total_signal)

    def __hash__(self):
        return hash(self.glycan_composition)

    def __str__(self):
        return str(self.glycan_composition)

    def __eq__(self, other):
        return self.glycan_composition == other

    def __repr__(self):
        return "BundledGlycanComposition(%s, %e)" % (self.glycan_composition, self.total_signal)

    @classmethod
    def aggregate(cls, observations):
        signals = Counter()
        for obs in observations:
            signals[obs.glycan_composition] += obs.total_signal
        return [cls(k, v) for k, v in signals.items() if v > 0]


class AggregatedAbundanceArtist(EntitySummaryBarChartArtist):
    y_label = "Relative Intensity"
    plot_title = "Glycan Composition\nTotal Abundances"

    def get_heights(self, items, logscale=False, *args, **kwargs):
        heights = [c.total_signal for c in items]
        if logscale:
            heights = np.log2(heights)
        return heights


class ScoreBarArtist(EntitySummaryBarChartArtist):
    y_label = "Composition Score"
    plot_title = "Glycan Composition Scores"

    def get_heights(self, items, *args, **kwargs):
        heights = [c.score for c in items]
        return heights



import itertools
from typing import Iterable, Optional
from uuid import uuid4
from collections import defaultdict, Counter
from multiprocessing import Process, Queue, Event, RLock
from threading import Thread
from itertools import product

from queue import Empty as QueueEmptyException

from glypy import Composition
from glypy.composition import formula
from glypy.structure.glycan_composition import FrozenGlycanComposition

from glycresoft.serialize import DatabaseBoundOperation, func
from glycresoft.serialize.hypothesis import GlycopeptideHypothesis
from glycresoft.serialize.hypothesis.peptide import Glycopeptide, Peptide, Protein, ProteinSite
from glycresoft.serialize.hypothesis.glycan import (
    GlycanCombination, GlycanClass, GlycanComposition,
    GlycanTypes, GlycanCombinationGlycanComposition)
from glycresoft.serialize.utils import toggle_indices
from glycresoft.task import TaskBase

from glycresoft.database.builder.glycan import glycan_combinator
from glycresoft.database.builder.base import HypothesisSerializerBase

from glycresoft.database.builder.glycopeptide.proteomics.remove_duplicate_peptides import DeduplicatePeptides

from glycopeptidepy.structure.sequence import (
    _n_glycosylation, _o_glycosylation, _gag_linker_glycosylation)

from glycopeptidepy.algorithm import reverse_sequence
from glycopeptidepy.structure.sequence import (
    find_n_glycosylation_sequons,
    PeptideSequence
)

from glycopeptidepy.structure.residue import UnknownAminoAcidException

_DEFAULT_GLYCAN_STEP_LIMIT = 15000


def slurp(session, model, ids, flatten=True):
    if flatten:
        ids = [j for i in ids for j in i]
    total = len(ids)
    last = 0
    step = 100
    results = []
    while last < total:
        results.extend(session.query(model).filter(
            model.id.in_(ids[last:last + step])))
        last += step
    return results


class GlycopeptideHypothesisSerializerBase(DatabaseBoundOperation, HypothesisSerializerBase):
    """Common machinery for Glycopeptide Hypothesis construction.

    Attributes
    ----------
    uuid : str
        The uuid of the hypothesis to be constructed
    """
    def __init__(self, database_connection, hypothesis_name=None, glycan_hypothesis_id=None,
                 full_cross_product=True, use_uniprot=True, uniprot_source_file: Optional[str]=None):
        DatabaseBoundOperation.__init__(self, database_connection)
        self._hypothesis_name = hypothesis_name
        self._hypothesis_id = None
        self._hypothesis = None
        self._glycan_hypothesis_id = glycan_hypothesis_id
        self.uuid = str(uuid4().hex)
        self.total_glycan_combination_count = -1
        self.use_uniprot = use_uniprot
        self.uniprot_source_file = uniprot_source_file
        self.full_cross_product = full_cross_product

    def _construct_hypothesis(self):
        if self._hypothesis_name is None or self._hypothesis_name.strip() == "":
            self._hypothesis_name = self._make_name()

        if self.glycan_hypothesis_id is None:
            raise ValueError("glycan_hypothesis_id must not be None")
        self._hypothesis = GlycopeptideHypothesis(
            name=self._hypothesis_name, glycan_hypothesis_id=self._glycan_hypothesis_id,
            uuid=self.uuid)
        self.session.add(self._hypothesis)
        self.session.commit()

        self._hypothesis_id = self._hypothesis.id
        self._hypothesis_name = self._hypothesis.name
        self._glycan_hypothesis_id = self._hypothesis.glycan_hypothesis_id

    def _make_name(self):
        return "GlycopeptideHypothesis-" + self.uuid

    @property
    def glycan_hypothesis_id(self):
        if self._glycan_hypothesis_id is None:
            self._construct_hypothesis()
        return self._glycan_hypothesis_id

    @property
    def n_glycan_only(self):
        if self._hypothesis is None:
            self._construct_hypothesis()
        return self._hypothesis.n_glycan_only

    def peptide_ids_with_n_glycosites(self):
        # May include the residue beyond the final
        q = self.session.query(Peptide.id.distinct()).join(Protein).join(Protein.sites).filter(
            Peptide.spans(ProteinSite.location) &
            (ProteinSite.name == ProteinSite.N_GLYCOSYLATION) &
            (Protein.hypothesis_id == self._hypothesis_id)).all()
        return [i[0] for i in q]

    def peptide_ids(self):
        if self.n_glycan_only:
            return self.peptide_ids_with_n_glycosites()
        q = self.session.query(Peptide.id).filter(Peptide.hypothesis_id == self._hypothesis_id).all()
        return [i[0] for i in q]

    def combinate_glycans(self, n):
        combinator = glycan_combinator.GlycanCombinationSerializer(
            self.engine, self.glycan_hypothesis_id,
            self.hypothesis_id, n)
        combinator.run()
        self.total_glycan_combination_count = combinator.total_count
        if not (self.total_glycan_combination_count > 0):
            raise ValueError("No glycan combinations were generated. No glycopeptides can be produced!")

    def _count_produced_glycopeptides(self):
        count = self.query(
            func.count(Glycopeptide.id)).filter(
            Glycopeptide.hypothesis_id == self.hypothesis_id).scalar()
        self.log("Generated %d glycopeptides" % count)
        self.set_parameters({
            "database_size": count
        })
        return count

    def _sql_analyze_database(self):
        self.log("Analyzing Indices")
        self._analyze_database()
        if self.is_sqlite():
            self._sqlite_reload_analysis_plan()
        self.log("Done Analyzing Indices")

    def deduplicate_peptides(self):
        DeduplicatePeptides(self._original_connection,
                            self.hypothesis_id).run()


class GlycopeptideHypothesisDestroyer(DatabaseBoundOperation, TaskBase):
    def __init__(self, database_connection, hypothesis_id):
        DatabaseBoundOperation.__init__(self, database_connection)
        self.hypothesis_id = hypothesis_id

    def delete_glycopeptides(self):
        self.log("Delete Glycopeptides")
        self.session.query(Glycopeptide).filter(
            Glycopeptide.hypothesis_id == self.hypothesis_id).delete(
            synchronize_session=False)
        self.session.commit()

    def delete_peptides(self):
        self.log("Delete Peptides")
        q = self.session.query(Protein.id).filter(Protein.hypothesis_id == self.hypothesis_id)
        for protein_id, in q:
            self.session.query(Peptide).filter(
                Peptide.protein_id == protein_id).delete(
                synchronize_session=False)
            self.session.commit()

    def delete_protein(self):
        self.log("Delete Protein")
        self.session.query(Protein).filter(Protein.hypothesis_id == self.hypothesis_id).delete(
            synchronize_session=False)
        self.session.commit()

    def delete_hypothesis(self):
        self.log("Delete Hypothesis")
        self.session.query(GlycopeptideHypothesis).filter(
            GlycopeptideHypothesis.id == self.hypothesis_id).delete()
        self.session.commit()

    def run(self):
        self.delete_glycopeptides()
        self.delete_peptides()
        self.delete_protein()
        self.delete_hypothesis()
        self.session.commit()


def distinct_glycan_classes(session, hypothesis_id):
    structure_classes = session.query(GlycanClass.name.distinct()).join(
        GlycanComposition.structure_classes).join(
        GlycanCombinationGlycanComposition).join(
        GlycanCombination).filter(
        GlycanCombination.hypothesis_id == hypothesis_id).all()
    return [sc[0] for sc in structure_classes]


def composition_to_structure_class_map(session, glycan_hypothesis_id):
    mapping = defaultdict(list)
    id_to_class_iterator = session.query(GlycanComposition.id, GlycanClass.name).join(
        GlycanComposition.structure_classes).filter(
        GlycanComposition.hypothesis_id == glycan_hypothesis_id).all()
    for gc_id, sc_name in id_to_class_iterator:
        mapping[gc_id].append(sc_name)
    return mapping


def combination_structure_class_map(session, hypothesis_id, composition_class_map):
    mapping = defaultdict(list)
    iterator = session.query(
        GlycanCombinationGlycanComposition).join(GlycanCombination).filter(
        GlycanCombination.hypothesis_id == hypothesis_id).order_by(GlycanCombination.id)
    for glycan_id, combination_id, count in iterator:
        listing = mapping[combination_id]
        for i in range(count):
            listing.append(composition_class_map[glycan_id])
    return mapping


class GlycanCombinationPartitionTable(TaskBase):
    def __init__(self, session, glycan_combinations, glycan_classes, hypothesis):
        self.session = session
        self.tables = defaultdict(lambda: defaultdict(list))
        self.hypothesis_id = hypothesis.id
        self.glycan_hypothesis_id = hypothesis.glycan_hypothesis_id
        self.glycan_classes = glycan_classes
        self.build_table(glycan_combinations)

    def build_table(self, glycan_combinations):
        composition_class_map = composition_to_structure_class_map(
            self.session, self.glycan_hypothesis_id)
        combination_class_map = combination_structure_class_map(
            self.session, self.hypothesis_id, composition_class_map)

        for entry in glycan_combinations:
            size_table = self.tables[entry.count]
            component_classes = combination_class_map[entry.id]
            class_assignment_generator = product(*component_classes)
            for classes in class_assignment_generator:
                counts = Counter(c for c in classes)
                key = tuple(counts[c] for c in self.glycan_classes)
                class_table = size_table[key]
                class_table.append(entry)

    def build_key(self, mapping):
        return tuple(mapping.get(c, 0) for c in self.glycan_classes)

    def get_entries(self, size, mapping):
        key = self.build_key(mapping)
        return self.tables[size][key]

    def __getitem__(self, key):
        size, mapping = key
        return self.get_entries(size, mapping)


def limiting_combinations(iterable: Iterable, n: int, limit: int=100):
    i = 0
    for result in itertools.combinations(iterable, n):
        i += 1
        yield result
        if i > limit:
            break


class GlycanCombinationRecord(object):
    __slots__ = [
        'id', 'calculated_mass', 'formula', 'count', 'glycan_composition_string',
        '_composition', '_dehydrated_composition']

    id: int
    calculated_mass: float
    formula: str
    count: int
    glycan_composition_string: str

    _composition: Composition
    _dehydrated_composition: Composition

    def __init__(self, combination):
        self.id = combination.id
        self.calculated_mass = combination.calculated_mass
        self.formula = combination.formula
        self.count = combination.count
        self.glycan_composition_string = combination.composition
        self._composition = None
        self._dehydrated_composition = None

    def total_composition(self):
        if self._composition is None:
            self._composition = self.convert().total_composition()
        return self._composition

    def dehydrated_composition(self):
        if self._dehydrated_composition is None:
            self._dehydrated_composition = self.total_composition() - (Composition("H2O") * self.count)
        return self._dehydrated_composition

    def convert(self):
        gc = FrozenGlycanComposition.parse(self.glycan_composition_string)
        gc.id = self.id
        gc.count = self.count
        return gc

    def __repr__(self):
        return "GlycanCombinationRecord(%d, %s)" % (
            self.id, self.glycan_composition_string)


class PeptideGlycosylator(object):
    def __init__(self, session, hypothesis_id, glycan_offset=None, glycan_limit=_DEFAULT_GLYCAN_STEP_LIMIT):
        self.session = session

        self.glycan_offset = glycan_offset
        self.glycan_limit = glycan_limit

        self.hypothesis_id = hypothesis_id
        self.hypothesis = self.session.query(GlycopeptideHypothesis).get(hypothesis_id)
        self.total_combinations = self._get_total_combination_count()

        self.build_glycan_table(self.glycan_offset)

    def _get_total_combination_count(self):
        count = self.session.query(
            GlycanCombination).filter(
            GlycanCombination.hypothesis_id == self.hypothesis_id).count()
        return count

    def _load_glycan_records(self):
        if self.glycan_offset is None:
            glycan_combinations = self.session.query(
                GlycanCombination).filter(
                GlycanCombination.hypothesis_id == self.hypothesis_id).all()
            glycan_combinations = [GlycanCombinationRecord(gc) for gc in glycan_combinations]
        else:
            glycan_combinations = self.session.query(
                GlycanCombination).filter(
                GlycanCombination.hypothesis_id == self.hypothesis_id).offset(
                self.glycan_offset).limit(self.glycan_limit).all()
        return glycan_combinations

    def _build_size_table(self, glycan_combinations):
        self.glycan_combination_partitions = GlycanCombinationPartitionTable(
            self.session, glycan_combinations, distinct_glycan_classes(
                self.session, self.hypothesis_id), self.hypothesis)

    def build_glycan_table(self, offset=None):
        self.glycan_offset = offset
        glycan_combinations = self._load_glycan_records()
        self._build_size_table(glycan_combinations)

    def handle_peptide(self, peptide):
        water = Composition("H2O")
        peptide_composition = Composition(str(peptide.formula))
        obj = peptide.convert()
        reference = obj.clone()

        # Handle N-linked glycosylation sites

        n_glycosylation_unoccupied_sites = set(peptide.n_glycosylation_sites)
        for site in list(n_glycosylation_unoccupied_sites):
            if obj[site][1]:
                n_glycosylation_unoccupied_sites.remove(site)
        for i in range(len(n_glycosylation_unoccupied_sites)):
            i += 1
            for gc in self.glycan_combination_partitions[i, {GlycanTypes.n_glycan: i}]:
                total_mass = peptide.calculated_mass + gc.calculated_mass - (gc.count * water.mass)
                formula_string = formula(peptide_composition + gc.dehydrated_composition())

                for site_set in limiting_combinations(n_glycosylation_unoccupied_sites, i):
                    sequence = reference.clone()
                    for site in site_set:
                        sequence.add_modification(site, _n_glycosylation.name)
                    sequence.glycan = gc.convert()

                    glycopeptide_sequence = str(sequence)

                    glycopeptide = dict(
                        calculated_mass=total_mass,
                        formula=formula_string,
                        glycopeptide_sequence=glycopeptide_sequence,
                        peptide_id=peptide.id,
                        protein_id=peptide.protein_id,
                        hypothesis_id=peptide.hypothesis_id,
                        glycan_combination_id=gc.id)
                    yield glycopeptide

        # Handle O-linked glycosylation sites
        o_glycosylation_unoccupied_sites = set(peptide.o_glycosylation_sites)
        for site in list(o_glycosylation_unoccupied_sites):
            if obj[site][1]:
                o_glycosylation_unoccupied_sites.remove(site)

        for i in range(len(o_glycosylation_unoccupied_sites)):
            i += 1
            for gc in self.glycan_combination_partitions[i, {GlycanTypes.o_glycan: i}]:
                total_mass = peptide.calculated_mass + gc.calculated_mass - (gc.count * water.mass)
                formula_string = formula(peptide_composition + gc.dehydrated_composition())

                for site_set in limiting_combinations(o_glycosylation_unoccupied_sites, i):
                    sequence = reference.clone()
                    for site in site_set:
                        sequence.add_modification(site, _o_glycosylation.name)
                    sequence.glycan = gc.convert()

                    glycopeptide_sequence = str(sequence)

                    glycopeptide = dict(
                        calculated_mass=total_mass,
                        formula=formula_string,
                        glycopeptide_sequence=glycopeptide_sequence,
                        peptide_id=peptide.id,
                        protein_id=peptide.protein_id,
                        hypothesis_id=peptide.hypothesis_id,
                        glycan_combination_id=gc.id)
                    yield glycopeptide

        # Handle GAG glycosylation sites
        gag_unoccupied_sites = set(peptide.gagylation_sites)
        for site in list(gag_unoccupied_sites):
            if obj[site][1]:
                gag_unoccupied_sites.remove(site)
        for i in range(len(gag_unoccupied_sites)):
            i += 1
            for gc in self.glycan_combination_partitions[i, {GlycanTypes.gag_linker: i}]:
                total_mass = peptide.calculated_mass + gc.calculated_mass - (gc.count * water.mass)
                formula_string = formula(peptide_composition + gc.dehydrated_composition())
                for site_set in limiting_combinations(gag_unoccupied_sites, i):
                    sequence = reference.clone()
                    for site in site_set:
                        sequence.add_modification(site, _gag_linker_glycosylation.name)
                    sequence.glycan = gc.convert()

                    glycopeptide_sequence = str(sequence)

                    glycopeptide = dict(
                        calculated_mass=total_mass,
                        formula=formula_string,
                        glycopeptide_sequence=glycopeptide_sequence,
                        peptide_id=peptide.id,
                        protein_id=peptide.protein_id,
                        hypothesis_id=peptide.hypothesis_id,
                        glycan_combination_id=gc.id)
                    yield glycopeptide


def null_log_handler(msg):
    print(msg)


class PeptideGlycosylatingProcess(Process):
    process_name = "glycopeptide-build-worker"

    def __init__(self, connection, hypothesis_id, input_queue, chunk_size=5000, done_event=None,
                 log_handler=null_log_handler, glycan_offset=None,
                 glycan_limit=_DEFAULT_GLYCAN_STEP_LIMIT):
        Process.__init__(self)
        self.daemon = True
        self.connection = connection
        self.input_queue = input_queue
        self.chunk_size = chunk_size
        self.hypothesis_id = hypothesis_id
        self.done_event = done_event
        self.log_handler = log_handler

        self.glycan_offset = glycan_offset
        self.glycan_limit = glycan_limit

        self.session = None
        self.work_done_event = Event()

    def is_work_done(self):
        return self.work_done_event.is_set()

    def process_result(self, collection):
        self.session.bulk_insert_mappings(Glycopeptide, collection, render_nulls=True)
        self.session.commit()

    def load_peptides(self, work_items):
        peptides = slurp(self.session, Peptide, work_items, flatten=False)
        return peptides

    def task(self):
        database = DatabaseBoundOperation(self.connection)
        self.session = database.session
        has_work = True

        glycosylator = PeptideGlycosylator(
            database.session, self.hypothesis_id,
            glycan_offset=self.glycan_offset,
            glycan_limit=self.glycan_limit)
        result_accumulator = []

        n = 0
        n_gps = 0
        while has_work:
            try:
                work_items = self.input_queue.get(timeout=5)
                if work_items is None:
                    has_work = False
                    continue
            except Exception:
                if self.done_event.is_set():
                    has_work = False
                continue
            peptides = self.load_peptides(work_items)
            n += len(peptides)
            for peptide in peptides:
                for gp in glycosylator.handle_peptide(peptide):
                    result_accumulator.append(gp)
                    if len(result_accumulator) > self.chunk_size:
                        n_gps += len(result_accumulator)
                        self.process_result(result_accumulator)
                        result_accumulator = []
            if len(result_accumulator) > 0:
                n_gps += len(result_accumulator)
                self.process_result(result_accumulator)
                result_accumulator = []
        self.work_done_event.set()
        # It seems there is no public API to force the process to check if it is done
        # but the internal method is invoked when creating a Process `repr` on Python 2.
        # This problem supposedly doesn't exist in Python 3.
        repr(self)
        self.log_handler("Process %r completed. (%d peptides, %d glycopeptides)" % (self.pid, n, n_gps))

    def run(self):
        new_name = getattr(self, 'process_name', None)
        if new_name is not None:
            TaskBase().try_set_process_name(new_name)
        try:
            self.task()
        except Exception as e:
            import traceback
            self.log_handler(
                "An exception has occurred for %r.\n%r\n%s" % (
                    self, e, traceback.format_exc()))


class NonSavingPeptideGlycosylatingProcess(PeptideGlycosylatingProcess):
    def process_result(self, collection):
        pass


class QueuePushingPeptideGlycosylatingProcess(PeptideGlycosylatingProcess):
    def __init__(self, connection, hypothesis_id, input_queue, output_queue, chunk_size=5000,
                 done_event=None, log_handler=null_log_handler, database_mutex=None,
                 glycan_offset=None, glycan_limit=_DEFAULT_GLYCAN_STEP_LIMIT):
        super(QueuePushingPeptideGlycosylatingProcess, self).__init__(
            connection, hypothesis_id, input_queue, chunk_size, done_event, log_handler,
            glycan_offset=glycan_offset, glycan_limit=glycan_limit)
        self.output_queue = output_queue
        self.database_mutex = database_mutex

    def load_peptides(self, work_items):
        with self.database_mutex:
            result = super(QueuePushingPeptideGlycosylatingProcess, self).load_peptides(work_items)
        return result

    def process_result(self, collection):
        self.output_queue.put(collection)


class MultipleProcessPeptideGlycosylator(TaskBase):
    def __init__(self, connection_specification, hypothesis_id, chunk_size=6500, n_processes=4,
                 glycan_combination_count=None, glycan_limit=_DEFAULT_GLYCAN_STEP_LIMIT):
        self.n_processes = n_processes
        self.connection_specification = connection_specification
        self.chunk_size = chunk_size
        self.hypothesis_id = hypothesis_id
        self.glycan_combination_count = glycan_combination_count
        self.current_glycan_offset = 0
        self.glycan_limit = glycan_limit

        self.input_queue = Queue(10)
        self.output_queue = Queue(1000)
        self.workers = []
        self.dealt_done_event = Event()
        self.ipc_controller = self.ipc_logger()
        self.database_mutex = RLock()

    def spawn_worker(self):
        worker = QueuePushingPeptideGlycosylatingProcess(
            self.connection_specification, self.hypothesis_id, self.input_queue,
            self.output_queue, self.chunk_size, self.dealt_done_event,
            self.ipc_controller.sender(), self.database_mutex,
            glycan_offset=self.current_glycan_offset,
            glycan_limit=self.glycan_limit)
        return worker

    def push_work_batches(self, peptide_ids):
        n = len(peptide_ids)
        i = 0
        chunk_size = max(min(int(n * 0.05), 1000), 5)
        while i < n:
            self.input_queue.put(peptide_ids[i:(i + chunk_size)])
            i += chunk_size
            self.log("... Dealt Peptides %d-%d %0.2f%%" % (i - chunk_size, min(i, n), (min(i, n) / float(n)) * 100))
        self.log("... All Peptides Dealt")
        self.dealt_done_event.set()

    def create_queue_feeder_thread(self, peptide_ids):
        queue_feeder = Thread(target=self.push_work_batches, args=(peptide_ids,))
        queue_feeder.daemon = True
        queue_feeder.start()
        return queue_feeder

    def spawn_all_workers(self):
        self.workers = []

        for i in range(self.n_processes):
            worker = self.spawn_worker()
            worker.start()
            self.workers.append(worker)

    def process(self, peptide_ids):
        connection = DatabaseBoundOperation(self.connection_specification)
        session = connection.session

        self.log("Begin Creation. Dropping Indices")
        index_controller = toggle_indices(session, Glycopeptide)
        index_controller.drop()

        while self.current_glycan_offset < self.glycan_combination_count:
            _current_progress = float(self.current_glycan_offset + self.glycan_limit)
            _current_percent_complete = _current_progress / self.glycan_combination_count * 100.0
            _current_percent_complete = min(_current_percent_complete, 100.0)
            self.log("... Processing Glycan Combinations %d-%d (%0.2f%%)" % (
                self.current_glycan_offset, min(self.current_glycan_offset + self.glycan_limit,
                                                self.glycan_combination_count),
                _current_percent_complete))
            queue_feeder = self.create_queue_feeder_thread(peptide_ids)
            self.spawn_all_workers()

            has_work = True
            last = 0
            i = 0
            while has_work:
                try:
                    batch = self.output_queue.get(True, 5)
                    try:
                        waiting_batches = self.output_queue.qsize()
                        if waiting_batches > 10:
                            with self.database_mutex:
                                self.log("... %d waiting sets." % (waiting_batches,))
                                try:
                                    for _ in range(waiting_batches):
                                        batch.extend(self.output_queue.get(True, 1))
                                    # check to see if any new work items have arrived while
                                    # we've been draining the queue
                                    waiting_batches = self.output_queue.qsize()
                                    if waiting_batches != 0:
                                        # if so, while the barrier is up, let's write the batch
                                        # to disk and then try to drain the queue again
                                        i += len(batch)
                                        try:
                                            session.bulk_insert_mappings(Glycopeptide, batch, render_nulls=True)
                                            session.commit()
                                        except Exception:
                                            session.rollback()
                                            raise
                                        batch = []
                                        for _ in range(waiting_batches):
                                            batch.extend(self.output_queue.get_nowait())
                                except QueueEmptyException:
                                    pass
                    except NotImplementedError:
                        # platform does not support qsize()
                        pass

                    with self.database_mutex:
                        i += len(batch)

                        try:
                            session.bulk_insert_mappings(Glycopeptide, batch, render_nulls=True)
                            session.commit()
                        except Exception:
                            session.rollback()
                            raise

                    if (i - last) > self.chunk_size * 20:
                        self.log("... %d Glycopeptides Created" % (i,))
                        last = i
                except QueueEmptyException:
                    if all(w.is_work_done() for w in self.workers):
                        has_work = False
                    continue

            queue_feeder.join()

            for worker in self.workers:
                self.log("Joining Process %r (%s)" % (worker.pid, worker.is_alive()))
                worker.join(10)
                if worker.is_alive():
                    self.log("Failed to join %r" % worker.pid)
                if worker.exitcode != 0 and worker.exitcode is not None:
                    raise ValueError("One or more workers failed to exit successfully")

            self.current_glycan_offset += self.glycan_limit

        self.log("All Work Done. Rebuilding Indices")
        index_controller.create()


class ProteinReversingMixin:
    def reverse_protein(self, protein: Protein) -> Protein:
        original_sequence = protein.protein_sequence
        n = len(original_sequence)
        if "(" in protein.protein_sequence:
            protein.protein_sequence = str(reverse_sequence(
                protein.protein_sequence, suffix_len=0))

        else:
            protein.protein_sequence = protein.protein_sequence[::-1]
        protein.hypothesis_id = self.hypothesis_id
        sites = []
        original_sequence = PeptideSequence(original_sequence)
        try:
            n_glycosites = find_n_glycosylation_sequons(
                original_sequence,
                include_cysteine=self.include_cysteine_n_glycosylation)
            for n_glycosite in n_glycosites:
                sites.append(
                    ProteinSite(name=ProteinSite.N_GLYCOSYLATION, location=n - n_glycosite - 1))
        except UnknownAminoAcidException:
            pass
        protein.sites.extend(sites)
        return protein



from typing import Optional, Union
from weakref import WeakValueDictionary

from ms_deisotope.data_source import ProcessedScan, PrecursorInformation, ActivationInformation, Scan
from ms_deisotope.data_source import ProcessedRandomAccessScanSource


class ScanStub(object):
    """A stub for holding precursor information and
    giving a Scan-like interface for accessing just that
    information. Provides a serialized-like interface
    which clients can use to load the real scan.

    Attributes
    ----------
    id : str
        The scan ID for the proxied scan
    precursor_information : :class:`~.PrecursorInformation`
        The information describing the relevant
        metadata for scheduling when and where this
        scan should be processed, where actual loading
        will occur.
    source : :class:`~.RandomAccessScanSource`
        A resource to use to load scans with by scan id.
    """

    id: str
    precursor_information: PrecursorInformation
    source: Optional[ProcessedRandomAccessScanSource]

    def __init__(self, precursor_information: PrecursorInformation, source: ProcessedRandomAccessScanSource):
        self.id = precursor_information.product_scan_id
        self.precursor_information = precursor_information
        self.source = source

    def detatch(self) -> 'ScanStub':
        self.source = None
        self.precursor_information.source = None
        return self

    def unbind(self):
        return self.detatch()

    def bind(self, source):
        self.source = source
        self.precursor_information.source = source
        return self

    @property
    def scan_id(self):
        return self.id

    def convert(self, *args, **kwargs) -> Union[ProcessedScan, Scan]:
        try:
            return self.source.get_scan_by_id(self.id)
        except AttributeError:
            raise KeyError(self.id)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.precursor_information.neutral_mass}, {self.precursor_information})"
        return template.format(self=self)


class ScanWrapperBase(object):
    __slots__ = []

    scan: Union[ProcessedScan, Scan]

    @property
    def scan_id(self):
        self.requires_scan()
        return self.scan.scan_id

    @property
    def precursor_ion_mass(self):
        self.requires_scan()
        neutral_mass = self.scan.precursor_information.extracted_neutral_mass
        if neutral_mass == 0:
            neutral_mass = self.scan.precursor_information.neutral_mass
        return neutral_mass

    @property
    def scan_time(self):
        self.requires_scan()
        return self.scan.scan_time

    @property
    def precursor_information(self):
        self.requires_scan()
        return self.scan.precursor_information

    def requires_scan(self):
        if self.scan is None:
            raise ValueError("%s is detatched from Scan" % (self.__class__.__name__))

    @property
    def ms_level(self):
        self.requires_scan()
        return self.scan.ms_level

    @property
    def index(self):
        self.requires_scan()
        return self.scan.index

    @property
    def activation(self):
        self.requires_scan()
        return self.scan.activation

    @property
    def deconvoluted_peak_set(self):
        self.requires_scan()
        return self.scan.deconvoluted_peak_set

    @property
    def peak_set(self):
        self.requires_scan()
        return self.scan.peak_set

    @property
    def arrays(self):
        self.requires_scan()
        return self.scan.arrays

    @property
    def annotations(self):
        self.requires_scan()
        return self.scan.annotations


class ScanInformation(object):
    """A carrier of scan-level metadata that does not include peak data to reduce space
    consumption.
    """

    __slots__ = ("id", "index", "scan_time", "ms_level",
                 "precursor_information", "activation",
                 "title", "__weakref__")

    id: str
    index: int
    scan_time: float
    ms_level: int
    precursor_information: PrecursorInformation
    activation: ActivationInformation
    title: str

    def __init__(self, scan_id, index, scan_time, ms_level, precursor_information, activation=None, title=None):
        self.id = scan_id
        self.index = index
        self.scan_time = scan_time
        self.ms_level = ms_level
        self.precursor_information = precursor_information
        self.activation = activation
        self.title = title

    def __reduce__(self):
        return self.__class__, (self.id, self.index, self.scan_time, self.ms_level,
                                self.precursor_information, self.activation, self.title)

    @property
    def scan_id(self):
        return self.id

    @classmethod
    def from_scan(cls, scan):
        return cls(
            scan.scan_id, scan.index, scan.scan_time, scan.ms_level,
            scan.precursor_information, scan.activation, scan.title)

    def __repr__(self):
        template = ("{self.__class__.__name__}({self.id}, {self.index}, {self.scan_time}, "
                    "{self.ms_level}, {self.precursor_information}, {self.activation}, {self.title})")
        return template.format(self=self)


class ScanInformationLoader(object):
    scan_loader: ProcessedRandomAccessScanSource

    def __init__(self, scan_loader):
        self.scan_loader = scan_loader
        self.cache = WeakValueDictionary()

    def get_scan_by_id(self, scan_id):
        try:
            return self.cache[scan_id]
        except KeyError:
            scan = self.scan_loader.get_scan_by_id(scan_id)
            info = ScanInformation.from_scan(scan)
            self.cache[scan_id] = info
            return info


def top_n_peaks(scan: ProcessedScan, n: int=300) -> ProcessedScan:
    scan = scan.clone(deep=True)
    peaks = scan.deconvoluted_peak_set.__class__(sorted(
        scan.deconvoluted_peak_set, key=lambda x: x.intensity, reverse=True)[:n])
    peaks.reindex()
    scan.deconvoluted_peak_set = peaks
    return scan



import math

from .base import (
    ChemicalShift, ModelTreeNode, EXDFragmentationStrategy,
    HCDFragmentationStrategy, IonSeries)
from .binomial_score import BinomialSpectrumMatcher, StubIgnoringBinomialSpectrumMatcher
from .simple_score import SimpleCoverageScorer, SignatureAwareCoverageScorer
from .precursor_mass_accuracy import MassAccuracyMixin


class CoverageWeightedBinomialScorer(BinomialSpectrumMatcher, SignatureAwareCoverageScorer, MassAccuracyMixin):

    def __init__(self, scan, sequence, mass_shift=None):
        super(CoverageWeightedBinomialScorer, self).__init__(scan, sequence, mass_shift)

    def _match_backbone_series(self, series, error_tolerance=2e-5, masked_peaks=None, strategy=None,
                               include_neutral_losses=False):
        if strategy is None:
            strategy = HCDFragmentationStrategy
        previous_position_glycosylated = False
        for frags in self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses):
            glycosylated_position = previous_position_glycosylated
            self.n_theoretical += 1
            for frag in frags:
                if not glycosylated_position:
                    glycosylated_position |= frag.is_glycosylated
                for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
                    if peak.index.neutral_mass in masked_peaks:
                        continue
                    self.solution_map.add(peak, frag)
            if glycosylated_position:
                if series.direction > 0:
                    self.glycosylated_n_term_ion_count += 1
                else:
                    self.glycosylated_c_term_ion_count += 1
            previous_position_glycosylated = glycosylated_position

    def match(self, error_tolerance=2e-5, *args, **kwargs):
        return SignatureAwareCoverageScorer.match(self, error_tolerance=error_tolerance, *args, **kwargs)

    def calculate_score(self, error_tolerance=2e-5, backbone_weight=None, glycosylated_weight=None,
                        stub_weight=None, *args, **kwargs):
        bin_score = BinomialSpectrumMatcher.calculate_score(
            self, error_tolerance=error_tolerance)
        coverage_score = SimpleCoverageScorer.calculate_score(
            self, backbone_weight, glycosylated_weight, stub_weight)
        mass_accuracy = self._precursor_mass_accuracy_score()
        signature_component = self._signature_ion_score()
        self._score = bin_score * coverage_score + mass_accuracy + signature_component
        return self._score


class ShortPeptideCoverageWeightedBinomialScorer(CoverageWeightedBinomialScorer):
    stub_weight = 0.65


def _short_peptide_test(scan, target, *args, **kwargs):
    return len(target) < 10


CoverageWeightedBinomialModelTree = ModelTreeNode(CoverageWeightedBinomialScorer, {
    _short_peptide_test: ModelTreeNode(ShortPeptideCoverageWeightedBinomialScorer, {}),
})


try:
    from glycresoft._c.tandem.tandem_scoring_helpers import CoverageWeightedBinomialScorer_match_backbone_series
    CoverageWeightedBinomialScorer._match_backbone_series = CoverageWeightedBinomialScorer_match_backbone_series
except ImportError:
    pass


class StubIgnoringCoverageWeightedBinomialScorer(CoverageWeightedBinomialScorer):
    def _sanitize_solution_map(self):
        san = list()
        for pair in self.solution_map:
            if pair.fragment.series not in ("oxonium_ion", "stub_glycopeptide"):
                san.append(pair)
        return san



from .search_space import (
    PeptideGlycosylator, GlycanCombinationRecord,
    StructureClassification, glycopeptide_key_t,
    GlycoformGeneratorBase, DynamicGlycopeptideSearchBase,
    PredictiveGlycopeptideSearch, IterativeGlycopeptideSearch,
    Record, Parser, IdKeyMaker)

from .multipart_fdr import (GlycopeptideFDREstimationStrategy, GlycopeptideFDREstimator)

from .workflow import (
    MultipartGlycopeptideIdentifier, PeptideDatabaseProxyLoader,
    make_memory_database_proxy_resolver, make_disk_backed_peptide_database)


__all__ = [
    "PeptideGlycosylator", "GlycanCombinationRecord",
    "StructureClassification", "glycopeptide_key_t",
    "GlycoformGeneratorBase", "DynamicGlycopeptideSearchBase",
    "PredictiveGlycopeptideSearch", "IterativeGlycopeptideSearch",
    "Record", "Parser", 'IdKeyMaker',

    "GlycopeptideFDREstimationStrategy", "GlycopeptideFDREstimator",

    "MultipartGlycopeptideIdentifier", "PeptideDatabaseProxyLoader",
    "make_memory_database_proxy_resolver", 'make_disk_backed_peptide_database',
]



import numbers as abc_numbers
from io import StringIO
from collections import defaultdict, deque
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np

from glypy.composition import composition_transform
from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, GlycanComposition


from glycopeptidepy import HashableGlycanComposition
from glycopeptidepy.structure.glycan import GlycanCompositionProxy

from glycresoft import symbolic_expression

from .space import (n_glycan_distance, composition_distance, CompositionSpace)
from .neighborhood import NeighborhoodCollection
from .rule import CompositionRuleClassifier


_hexose = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
_hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")


class CompositionNormalizer(object):
    cache: Dict[HashableGlycanComposition, HashableGlycanComposition]

    def __init__(self, cache=None):
        if cache is None:
            cache = dict()
        self.cache = cache

    def _normalize_key(self, key: Union[str, GlycanCompositionProxy, GlycanComposition]) -> HashableGlycanComposition:
        if isinstance(key, str):
            key = HashableGlycanComposition.parse(key)
        return HashableGlycanComposition(
            {self._normalize_monosaccharide(k): v for k, v in key.items()})

    def _normalize_monosaccharide(self, key):
        is_derivatized = composition_transform.has_derivatization(key)
        if is_derivatized:
            key = key.copy_underivatized()
        return key

    def _get_solution(self, key: Union[HashableGlycanComposition, str,
                                       GlycanComposition, GlycanCompositionProxy]) -> HashableGlycanComposition:
        if isinstance(key, (str, HashableGlycanComposition)):
            try:
                return self.cache[key]
            except KeyError:
                result = self._normalize_key(key)
                self.cache[key] = result
                return result
        else:
            return self._normalize_key(key)

    def normalize_composition(self, c):
        return self._get_solution(c)

    def __call__(self, c):
        return self.normalize_composition(c)

    def copy(self):
        return self.__class__(self.cache.copy())


normalize_composition = CompositionNormalizer()


class DijkstraPathFinder(object):

    def __init__(self, graph, start, end, limit=float('inf')):
        self.graph = graph
        self.start = start
        self.end = end
        self.distance = defaultdict(lambda: float('inf'))
        self.distance[start._str] = 0
        self.unvisited_finite_distance = dict()
        self.limit = limit

    def find_smallest_unvisited(self, unvisited):
        smallest_distance = float('inf')
        smallest_node = None

        if not self.unvisited_finite_distance:
            iterable = [(k, self.distance[k]) for k in unvisited]
        else:
            iterable = self.unvisited_finite_distance.items()

        for node, distance in iterable:
            if distance <= smallest_distance:
                smallest_distance = distance
                smallest_node = node
        return self.graph[smallest_node]

    def find_path(self):
        unvisited = set([node._str for node in self.graph])

        visit_queue = deque([self.start])
        while self.end._str in unvisited:
            try:
                current_node = visit_queue.popleft()
            except IndexError:
                current_node = self.find_smallest_unvisited(unvisited)
            try:
                unvisited.remove(current_node._str)
            except KeyError:
                continue
            try:
                self.unvisited_finite_distance.pop(current_node._str)
            except KeyError:
                pass
            edges = current_node.edges
            for edge in edges:
                terminal = edge._traverse(current_node)
                if terminal._str not in unvisited:
                    continue
                path_length = self.distance[current_node._str] + edge.order
                terminal_distance = self.distance[terminal._str]
                if terminal_distance > path_length:
                    self.distance[terminal._str] = path_length
                    if terminal._str in unvisited and terminal_distance < self.limit:
                        self.unvisited_finite_distance[
                            terminal._str] = path_length
                if terminal_distance < self.limit:
                    visit_queue.append(terminal)

    def search(self):
        self.find_path()
        return self.distance[self.end._str]


try:
    _has_c = True
    from glycresoft._c.composition_network.graph import DijkstraPathFinder
except ImportError:
    _has_c = False

class CompositionGraphNode(object):
    composition: HashableGlycanComposition
    index: int
    _score: float
    marked: bool
    edges: 'EdgeSet'
    _str: str
    _hash: int
    internal_score: float

    __slots__ = (
        "composition", "index", "_score", "marked", "edges", "_str", "_hash", "internal_score")

    def __init__(self, composition, index, score=0., marked=False, **kwargs):
        self.composition = composition
        self.index = index
        self.edges = EdgeSet()
        self._str = str(self.composition)
        self._hash = hash(str(self._str))
        self._score = score
        self.internal_score = 0.0
        self.marked = marked

    @property
    def glycan_composition(self):
        return self.composition

    @property
    def order(self):
        return len(self.edges)

    @property
    def score(self):
        if self._score == 0:
            return self.internal_score
        else:
            return self._score

    @score.setter
    def score(self, value):
        self._score = value

    def edge_to(self, node):
        return self.edges.edge_to(self, node)

    def neighbors(self):
        result = [edge._traverse(self) for edge in self.edges]
        return result

    def __eq__(self, other):
        try:
            return (self)._str == str(other)
        except AttributeError:
            return str(self) == str(other)

    def __str__(self):
        return self._str

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return "CompositionGraphNode(%s, %d, %0.2f)" % (
            self._str, int(self.index) if self.index is not None else -1,
            self.score if self.score != 0 else self.internal_score)

    def copy(self):
        dup = CompositionGraphNode(self.composition, self.index, self.score)
        dup.internal_score = self.internal_score
        return dup

    def clone(self):
        return self.copy()


class EdgeSet(object):
    store: Dict[Tuple[CompositionGraphNode, CompositionGraphNode], 'CompositionGraphEdge']

    def __init__(self, store=None):
        if store is None:
            store = dict()
        self.store = store

    def edge_to(self, node1, node2):
        if node2.index < node1.index:
            node1, node2 = node2, node1
        return self.store[node1, node2]

    def add(self, edge):
        self.store[edge.node1, edge.node2] = edge

    def add_if_shorter(self, edge):
        try:
            prev = self.store[edge.node1, edge.node2]
            if prev.order < edge.order:
                return False
            else:
                self.store[edge.node1, edge.node2] = edge
                return True
        except KeyError:
            self.store[edge.node1, edge.node2] = edge
            return True

    def remove(self, edge):
        self.store.pop((edge.node1, edge.node2))

    def __iter__(self):
        return iter(self.store.values())

    def __len__(self):
        return len(self.store)

    def __repr__(self):
        return str(set(self.store.values()))

    def __eq__(self, other):
        return self.store == other.store


class CompositionGraphEdge(object):
    __slots__ = ["node1", "node2", "order", "weight", "_hash", "_str"]

    node1: CompositionGraphNode
    node2: CompositionGraphNode
    order: int
    weight: float
    _hash: int
    _str: str

    def __init__(self, node1, node2, order, weight=1.0):
        self.node1 = node1
        self.node2 = node2
        self.order = order if order > 0 else 1
        self.weight = weight
        self._hash = hash((node1, node2, order))
        self._str = "(%s)" % ', '.join(map(str, (node1, node2, order)))

        node1.edges.add_if_shorter(self)
        node2.edges.add_if_shorter(self)

    def __getitem__(self, key):
        str_key = str(key)
        if str_key == self.node1._str:
            return self.node2
        elif str_key == self.node2._str:
            return self.node1
        else:
            raise KeyError(key)

    def _traverse(self, node):
        return self.node1 if node is self.node2 else self.node2

    def __eq__(self, other):
        return self._str == other._str

    def __str__(self):
        return self._str

    def __repr__(self):
        return self._str

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return self.__class__, (self.node1, self.node2, self.order, self.weight)

    def copy_for(self, node1, node2):
        return self.__class__(node1, node2, self.order, self.weight)

    def remove(self):
        try:
            self.node1.edges.remove(self)
        except KeyError:
            pass
        try:
            self.node2.edges.remove(self)
        except KeyError:
            pass


class CompositionGraphBase(object):
    def copy(self):
        graph = CompositionGraph([], self.distance_fn)
        graph._composition_normalizer = self._composition_normalizer.copy()
        for node in self.nodes:
            graph.add_node(node.clone())
        for edge in self.edges:
            n1 = graph.nodes[edge.node1.index]
            n2 = graph.nodes[edge.node2.index]
            e = edge.copy_for(n1, n2)
            graph.edges.add(e)
        graph.neighborhoods.update(self.neighborhoods.copy())
        return graph


try:
    _has_c = True
    from glycresoft._c.composition_network.graph import (
        CompositionGraphEdge, CompositionGraphNode, EdgeSet, CompositionGraphBase)
except ImportError:
    _has_c = False


class CompositionGraph(CompositionGraphBase):
    neighborhoods: NeighborhoodCollection
    nodes: List[CompositionGraphNode]
    node_map: Dict[HashableGlycanComposition, CompositionGraphNode]
    _composition_normalizer: CompositionNormalizer
    distance_fn: Callable[[HashableGlycanComposition, HashableGlycanComposition], Tuple[float, float]]
    edges: EdgeSet
    cache_state: Dict

    def __init__(self, compositions, distance_fn=n_glycan_distance, neighborhoods=None, cache_state: Optional[Dict]=None):
        if neighborhoods is None:
            neighborhoods = []
        if cache_state is None:
            cache_state = {}
        self.nodes = []
        self.node_map = {}
        self._composition_normalizer = CompositionNormalizer()
        self.distance_fn = distance_fn
        self.create_nodes(compositions)
        self.edges = EdgeSet()
        self.neighborhoods = NeighborhoodCollection(neighborhoods)
        self.cache_state = cache_state

    def create_nodes(self, compositions):
        """Given an iterable of GlycanComposition-like or strings encoding GlycanComposition-like
        objects construct nodes representing these compositions and add them to the graph.

        The order items appear in the list will affect their graph node's :attr:`index` attribute
        and in turn affect their order in edges. For consistent behavior, order nodes by mass.

        Parameters
        ----------
        compositions : Iterable
            An iterable source of GlycanComposition-like objects
        """
        i = 0
        compositions = map(self.normalize_composition, compositions)

        for c in sorted(set(compositions), key=lambda x: (x.mass(), len(x))):
            n = CompositionGraphNode(c, i)
            self.add_node(n)
            i += 1

    def add_node(self, node: CompositionGraphNode, reindex=False):
        """Given a CompositionGraphNode, add it to the graph.

        If `reindex` is `True` then a full re-indexing of the graph will
        take place after the insertion is made. Otherwise it is assumed that
        `node` is being added in index-specified order.

        Parameters
        ----------
        node : CompositionGraphNode
            The node to be added
        """
        key = str(self.normalize_composition(node.composition))
        if key in self.node_map:
            raise ValueError("Redundant composition!")
        self.nodes.append(node)
        self.node_map[key] = node

        if reindex:
            self._reindex()
        return self

    def create_edges(self, distance=1, distance_fn=composition_distance):
        """Traverse composition-space to find nodes similar to each other
        and construct CompositionGraphEdges between them to represent that
        similarity.

        Parameters
        ----------
        distance : int, optional
            The maximum dissimilarity between two nodes permitted
            to allow the construction of an edge between them
        distance_fn : callable, optional
            A function to use to compute the bounded distance between two nodes
            that are within `distance` of eachother in raw composition-space
        """
        if distance_fn is None:
            distance_fn = self.distance_fn
        space = CompositionSpace([node.composition for node in self])
        for node in self:
            if node is None:
                continue
            for related in space.find_related(node.composition, distance):
                related_node = self.node_map[str(related)]
                # Ensures no duplicates assuming symmetric search window
                if node.index < related_node.index:
                    diff, weight = distance_fn(node.composition, related)
                    e = CompositionGraphEdge(node, related_node, diff, weight)
                    self.edges.add(e)
        return self

    def add_edge(self, node1: CompositionGraphNode, node2: CompositionGraphNode):
        if node1.index > node2.index:
            node1, node2 = node2, node1
        diff, weight = self.distance_fn(node1.composition, node2.composition)
        e = CompositionGraphEdge(node1, node2, diff, weight)
        self.edges.add(e)

    def remove_node(self, node: CompositionGraphNode, bridge: bool=True, limit: int=5, ignore_marked: bool=True):
        """Removes the Glycan Composition given by `node` from the graph
        and all edges connecting to it. This will reindex the graph.

        If two Glycan Compositions `x` and `y` are connected *through* `node`,
        `bridge` is true,  and the shortest path connecting `x` and
        `y` is longer than the sum of the path from `x` to `node` and from `node`
        to `y`, create a new edge connecting `x` and `y`, if the new edge is shorter
        than `limit`.

        Parameters
        ----------
        node : CompositionGraphNode-like
            The node to be removed
        bridge : bool, optional
            Whether or not to create new edges
            bridging neighbors
        limit : int, optional
            The maximum path length under which
            to bridge neighbors

        Returns
        -------
        list
            The list of edges removed
        """
        node = self[node.glycan_composition]
        subtracted_edges = list(node.edges)
        for edge in subtracted_edges:
            self.remove_edge(edge)
        self.nodes.pop(node.index)
        self.node_map.pop(str(node.composition))
        self._reindex()

        if bridge:
            seen = set()
            for edge in subtracted_edges:
                for other_edge in subtracted_edges:
                    if edge == other_edge:
                        continue
                    node1 = edge._traverse(node)
                    node2 = other_edge._traverse(node)
                    if ignore_marked and (node1.marked or node2.marked):
                        continue
                    key = frozenset((node1.index, node2.index))
                    if key in seen:
                        continue
                    seen.add(key)
                    old_distance = edge.order + other_edge.order
                    if old_distance >= limit:
                        continue
                    try:
                        if node1.edge_to(node2):
                            continue
                    except KeyError:
                        pass
                    path_finder = DijkstraPathFinder(
                        self, node1, node2, min(old_distance + 1, limit))
                    shortest_path = path_finder.search()

                    # The distance function may not be transitive, in which case, an "edge through"
                    # solution may not make sense.
                    # shortest_path = self.distance_fn(node1._composition, node2._composition)[0]
                    path_length = min(old_distance, shortest_path)
                    if path_length < limit:
                        if node1.index > node2.index:
                            node1, node2 = node2, node1
                        new_edge = CompositionGraphEdge(
                            node1, node2, path_length, 1.)
                        self.edges.add_if_shorter(new_edge)
        return subtracted_edges

    def remove_edge(self, edge: CompositionGraphEdge):
        edge.remove()
        self.edges.remove(edge)

    def _reindex(self):
        i = 0
        for node in self:
            node.index = i
            i += 1

    def normalize_composition(self, composition):
        return self._composition_normalizer(composition)

    def __getitem__(self, key):
        if isinstance(key, (str, GlycanComposition, GlycanCompositionProxy)):
            key = self.normalize_composition(key)
            return self.node_map[key]
        # use the ABC Integral to catch all numerical types that could be used as an
        # index including builtin int and long, as well as all the NumPy flavors of
        # integers
        elif isinstance(key, (abc_numbers.Integral, slice)):
            return self.nodes[key]
        else:
            try:
                result = []
                iter(key)
                try:
                    if isinstance(key[0], (bool, np.bool_)):
                        for b, k in zip(key, self):
                            if b:
                                result.append(k)
                    else:
                        for k in key:
                            result.append(self[k])
                except TypeError:
                    # Handle unindexable iteratables
                    for k in key:
                        result.append(self[k])
                return result
            except Exception as e:
                if len(key) == 0:
                    return []
                raise IndexError(
                    "An error occurred (%r) during indexing with %r" % (e, key))

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    def __reduce__(self):
        return self.__class__, ([], self.distance_fn), self.__getstate__()

    def __getstate__(self):
        string_buffer = StringIO()
        dump(self, string_buffer)
        return string_buffer, self.distance_fn

    def __setstate__(self, state):
        string_buffer, distance_fn = state
        string_buffer.seek(0)
        net, neighborhoods = load(string_buffer)
        self.nodes = net.nodes
        self.neighborhoods = NeighborhoodCollection(neighborhoods)
        self.node_map = net.node_map
        self.edges = net.edges
        self.distance_fn = distance_fn
        self._composition_normalizer = CompositionNormalizer()

    def __repr__(self):
        return "{self.__class__.__name__}({node_count} nodes, {edge_count} edges)".format(
            self=self, node_count=len(self), edge_count=len(self.edges))

    def clone(self):
        return self.copy()

    def __eq__(self, other):
        try:
            return self.nodes == other.nodes and self.edges == other.edges
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def assign(self, observed, inplace=False):
        if not inplace:
            network = self.clone()
        else:
            network = self
        solution_map = {}
        for case in observed:
            if case.glycan_composition is None:
                continue
            s = solution_map.get(case.glycan_composition)
            if s is None or s.score < case.score:
                solution_map[case.glycan_composition] = case

        for node in network.nodes:
            node.internal_score = 0

        for composition, solution in solution_map.items():
            try:
                node = network[composition]
                node.internal_score = solution.internal_score
            except KeyError:
                # Not all exact compositions have nodes
                continue
        return network

    def unassign(self, inplace=False):
        if not inplace:
            network = self.clone()
        else:
            network = self
        for node in network:
            node.score = node.internal_score = 0
        return network

    def augment_with_decoys(self, pseudodistance=2):
        compositions = []
        if pseudodistance == 0:
            raise ValueError("Cannot use decoys with pseudodistance of size 0")
        for node in self:
            t = node.glycan_composition.copy()
            d = node.glycan_composition.copy()
            d['#decoy#'] = pseudodistance
            compositions.append(t)
            compositions.append(d)
        return self.__class__(compositions, self.distance_fn, self.neighborhoods)

    def merge(self, other):
        edges = other.edges
        for node in other:
            self.add_node(node)
        self._reindex()
        for edge in edges:
            self.add_edge(self[edge.node1], self[edge.node2])
        return self


try:
    _has_c = True
    from glycresoft._c.composition_network.graph import reindex_graph as _reindex_graph
    CompositionGraph._reindex = _reindex_graph
except ImportError:
    _has_c = False


class GraphWriter(object):

    def __init__(self, network, file_obj):
        self.network = network
        self.file_obj = file_obj
        self.handle_graph(self.network)

    def write(self, text):
        self.file_obj.write(text)

    def handle_node(self, node):
        composition = node.composition.serialize()
        index = (node.index)
        score = node.score
        self.write("%d\t%s\t%f\n" % (index, composition, score))

    def handle_edge(self, edge):
        index1 = edge.node1.index
        index2 = edge.node2.index
        diff = edge.order
        weight = edge.weight
        line = ("%d\t%d\t%d\t%f" % (index1, index2, diff, weight))
        trimmed = line.rstrip("0")
        if line != trimmed:
            trimmed += '0'
        self.write(trimmed + '\n')

    def handle_graph(self, graph):
        self.write("#COMPOSITIONGRAPH 1.1\n")
        self.write("#NODE\n")
        for node in self.network.nodes:
            self.handle_node(node)
        self.write("#EDGE\n")
        for edge in self.network.edges:
            self.handle_edge(edge)
        try:
            if graph.neighborhoods:
                self.write("#NEIGHBORHOOD\n")
                for neighborhood in graph.neighborhoods:
                    self.handle_neighborhood(neighborhood)
        except AttributeError:
            pass

    def handle_neighborhood(self, neighborhood_rule):
        self.write("BEGIN NEIGHBORHOOD\n")
        self.write(neighborhood_rule.serialize())
        self.write("END NEIGHBORHOOD\n")


class GraphReader(object):

    @classmethod
    def read(cls, file_obj):
        inst = cls(file_obj)
        return inst.network, inst.neighborhoods

    def __init__(self, file_obj):
        self.file_obj = file_obj
        self.network = CompositionGraph([])
        self.neighborhoods = self.network.neighborhoods
        self.handle_graph_file()

    def handle_node_line(self, line):
        try:
            index, composition, score = line.replace("\n", "").split("\t")
            score = float(score)
        except ValueError:
            index, composition = line.replace("\n", "").split("\t")
            score = 0.

        index = int(index)
        composition = HashableGlycanComposition.parse(composition)
        node = CompositionGraphNode(composition, index, score)
        node.internal_score = score
        self.network.nodes.append(node)
        self.network.node_map[str(node.composition)] = node

    def handle_edge_line(self, line):
        index1, index2, diff, weight = line.replace("\n", "").split("\t")
        index1, index2, diff, weight = int(index1), int(
            index2), int(diff), float(weight)
        node1 = self.network.nodes[index1]
        node2 = self.network.nodes[index2]
        edge = CompositionGraphEdge(node1, node2, diff, weight)
        self.network.edges.add(edge)

    def handle_graph_file(self):
        state = "START"
        buffering = []
        for line in self.file_obj:
            line = line.strip()
            if state == "START":
                if line == "#NODE":
                    state = "NODE"
            elif state == "NODE":
                if line == "#EDGE":
                    state = "EDGE"
                else:
                    self.handle_node_line(line)
            elif state == "EDGE":
                if line == "#NEIGHBORHOOD":
                    state = "NEIGHBORHOOD"
                else:
                    self.handle_edge_line(line)
            elif state == "NEIGHBORHOOD":
                if line.startswith("BEGIN NEIGHBORHOOD"):
                    if buffering:
                        self.handle_neighborhood(buffering)
                    buffering = []
                elif line.startswith("END NEIGHBORHOOD"):
                    if buffering:
                        self.handle_neighborhood(buffering)
                        buffering = []
                else:
                    buffering.append(line)

    def handle_neighborhood(self, lines):
        rule = CompositionRuleClassifier.parse(lines)
        self.neighborhoods.add(rule)


dump = GraphWriter
load = GraphReader.read



import unittest

import glypy
import pickle

from textwrap import dedent
from io import BytesIO

from glycresoft.database import composition_network
from glycresoft.database.builder.glycan import constrained_combinatorics

compositions = [
    glypy.glycan_composition.FrozenGlycanComposition(HexNAc=2, Hex=i) for i in range(3, 10)
] + [
    glypy.glycan_composition.FrozenGlycanComposition(HexNAc=4, Hex=5, NeuAc=2),
    glypy.glycan_composition.FrozenGlycanComposition(HexNAc=3, Hex=5),
]


_PERMETHYLATED_COMPOSITIONS = '''{Fuc^Me:1; Hex^Me:6; HexNAc^Me:5; Neu5NAc^Me:3}$C1H4
{Fuc^Me:2; Hex^Me:6; HexNAc^Me:5}$C1H4
{Fuc^Me:2; Hex^Me:6; HexNAc^Me:5; Neu5NAc^Me:2}$C1H4
{Fuc^Me:2; Hex^Me:6; HexNAc^Me:5; Neu5NAc^Me:3}$C1H4
{Fuc^Me:3; Hex^Me:6; HexNAc^Me:5; Neu5NAc^Me:2}$C1H4
{Fuc^Me:1; Hex^Me:3; HexNAc^Me:5}$C1H4
{Fuc^Me:2; Hex^Me:7; HexNAc^Me:8}$C1H4
{Hex^Me:6; HexNAc^Me:8; Neu5NAc^Me:1}$C1H4
{Fuc^Me:1; Hex^Me:3; HexNAc^Me:3}$C1H4
{Hex^Me:8; HexNAc^Me:3; Neu5NAc^Me:1}$C1H4
{Fuc^Me:1; Hex^Me:8; HexNAc^Me:3; Neu5NAc^Me:1}$C1H4
{Fuc^Me:2; Hex^Me:7; HexNAc^Me:4}$C1H4
{Fuc^Me:3; Hex^Me:7; HexNAc^Me:4}$C1H4
{Hex^Me:7; HexNAc^Me:5; Neu5NAc^Me:3}$C1H4
{Fuc^Me:1; Hex^Me:7; HexNAc^Me:5; Neu5NAc^Me:2}$C1H4
{Fuc^Me:2; Hex^Me:7; HexNAc^Me:5; Neu5NAc^Me:2}$C1H4
{Fuc^Me:1; Hex^Me:4; HexNAc^Me:4}$C1H4
{Fuc^Me:3; Hex^Me:4; HexNAc^Me:4; Neu5NAc^Me:1}$C1H4
{Fuc^Me:3; Hex^Me:4; HexNAc^Me:6; Neu5NAc^Me:1}$C1H4
{Hex^Me:5; HexNAc^Me:4}$C1H4
{Hex^Me:5; HexNAc^Me:4; Neu5NAc^Me:2}$C1H4
{Fuc^Me:1; Hex^Me:5; HexNAc^Me:4; Neu5NAc^Me:1}$C1H4
{Fuc^Me:1; Hex^Me:5; HexNAc^Me:4; Neu5NAc^Me:2}$C1H4
{Hex^Me:5; HexNAc^Me:5; Neu5NAc^Me:1}$C1H4
{Fuc^Me:1; Hex^Me:5; HexNAc^Me:5; Neu5NAc^Me:1}$C1H4
{Fuc^Me:1; Hex^Me:5; HexNAc^Me:5; Neu5NAc^Me:3}$C1H4
{Hex^Me:5; HexNAc^Me:7}$C1H4
{Fuc^Me:1; Hex^Me:5; HexNAc^Me:7}$C1H4
{Fuc^Me:1; Hex^Me:6; HexNAc^Me:4; Neu5NAc^Me:1}$C1H4
{Fuc^Me:2; Hex^Me:6; HexNAc^Me:4; Neu5NAc^Me:1}$C1H4
{Hex^Me:6; HexNAc^Me:5; Neu5NAc^Me:3}$C1H4
'''.splitlines()


class NeighborhoodWalkerTest(unittest.TestCase):

    def make_human_definition_buffer(self):
        text = b'''Hex 3 10
HexNAc 2 9
Fuc 0 5
NeuAc 0 5

Fuc < HexNAc
HexNAc > NeuAc + 1'''
        buff = BytesIO()
        buff.write(text)
        buff.seek(0)
        return buff

    def make_mammalian_definition_buffer(self):
        text = b'''Hex 3 10
HexNAc 2 9
Fuc 0 5
NeuAc 0 5
NeuGc 0 5

Fuc < HexNAc
HexNAc > (NeuAc + NeuGc) + 1'''
        buff = BytesIO()
        buff.write(text)
        buff.seek(0)
        return buff

    def generate_compositions(self, rule_buffer):
        rules, constraints = constrained_combinatorics.parse_rules_from_file(
            rule_buffer)
        compositions = list(constrained_combinatorics.CombinatoricCompositionGenerator(
            rules_table=rules, constraints=constraints))
        # strip out glycan classes
        compositions = [c[0] for c in compositions]
        return compositions

    def test_human_network(self):
        rules_buffer = self.make_human_definition_buffer()
        compositions = self.generate_compositions(rules_buffer)
        g = composition_network.CompositionGraph(compositions)
        self.assertEqual(len(g), 1424)
        walker = composition_network.NeighborhoodWalker(g)

        neighborhood_sizes = {
            "tri-antennary": 188,
            "bi-antennary": 104,
            "asialo-bi-antennary": 96,
            "tetra-antennary": 276,
            "hybrid": 80,
            # "over-extended": 170,
            "asialo-tri-antennary": 60,
            "penta-antennary": 336,
            "asialo-penta-antennary": 72,
            "high-mannose": 16,
            "asialo-tetra-antennary": 68}
        for k, v in neighborhood_sizes.items():
            self.assertEqual(v, len(walker.neighborhood_maps[k]), "%s had %d members, not %d" % (
                k, len(walker.neighborhood_maps[k]), v))

    def test_mammalian_network(self):
        rules_buffer = self.make_mammalian_definition_buffer()
        compositions = self.generate_compositions(rules_buffer)
        g = composition_network.CompositionGraph(compositions)
        self.assertEqual(len(g), 5096)

        walker = composition_network.NeighborhoodWalker(g)

        neighborhood_sizes = {
            'tri-antennary': 408,
            'bi-antennary': 180,
            'asialo-bi-antennary': 144,
            'tetra-antennary': 720,
            'hybrid': 140,
            # 'over-extended': 363,
            'asialo-tri-antennary': 60,
            'penta-antennary': 1080,
            'asialo-penta-antennary': 72,
            'high-mannose': 16,
            'asialo-tetra-antennary': 68,
        }
        for k, v in neighborhood_sizes.items():
            self.assertEqual(v, len(walker.neighborhood_maps[k]), "%s had %d members, not %d" % (
                k, len(walker.neighborhood_maps[k]), v))


class CompositionGraphTest(unittest.TestCase):

    def test_construction(self):
        g = composition_network.CompositionGraph(compositions)
        self.assertTrue(len(g) == 9)
        self.assertTrue(len(g.edges) == 0)
        g.create_edges(1)
        self.assertTrue(len(g.edges) == 7)

    def test_index(self):
        g = composition_network.CompositionGraph(compositions)
        node = g["{Hex:5; HexNAc:2}"]
        self.assertTrue(node.glycan_composition["Hex"] == 5)
        self.assertTrue(node.glycan_composition["HexNAc"] == 2)
        i = node.index
        self.assertEqual(g[i], node)

    def test_bridge(self):
        g = composition_network.CompositionGraph(compositions)
        g.create_edges(1)
        for edge in g.edges:
            self.assertTrue(edge.order < 2)
        node_to_remove = g["{Hex:5; HexNAc:2}"]
        removed_edges = g.remove_node(node_to_remove)
        neighbors = [e[node_to_remove] for e in removed_edges]
        assert len(neighbors) == 3
        n_order_2_edges = 0
        for edge in g.edges:
            if edge.order == 2:
                self.assertTrue(
                    edge.node1 in neighbors and edge.node2 in neighbors)
                n_order_2_edges += 1
        self.assertTrue(n_order_2_edges > 0)
        node_to_remove = g["{Hex:6; HexNAc:2}"]
        removed_edges = g.remove_node(node_to_remove)
        neighbors = [e[node_to_remove] for e in removed_edges]
        assert len(neighbors) == 3
        for edge in g.edges:
            if edge.order == 3:
                self.assertTrue(
                    edge.node1 in neighbors and edge.node2 in neighbors)

    def test_pickle(self):
        g = composition_network.CompositionGraph(compositions)
        g.create_edges(1)
        self.assertEqual(g, pickle.loads(pickle.dumps(g)))

    def test_clone(self):
        g = composition_network.CompositionGraph(compositions)
        g.create_edges(1)
        self.assertEqual(g, g.clone())

    def test_normalize(self):
        g = composition_network.CompositionGraph(_PERMETHYLATED_COMPOSITIONS)
        self.assertIsNotNone(g["{Fuc:1; Hex:6; HexNAc:5; Neu5Ac:3}"])
        self.assertEqual(g["{Fuc^Me:1; Hex^Me:6; HexNAc^Me:5; Neu5Ac^Me:3}$C1H4"], g["{Fuc:1; Hex:6; HexNAc:5; Neu5Ac:3}"])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from collections import namedtuple, OrderedDict
from typing import List, Tuple

from six import string_types as basestring

import numpy as np
from scipy import linalg
from matplotlib import pyplot as plt

from glypy import GlycanComposition
from glycopeptidepy.structure.glycan import GlycanCompositionProxy

from glycresoft.database.composition_network import (
    NeighborhoodWalker, CompositionGraphNode, CompositionGraph)

from .constants import DEFAULT_LAPLACIAN_REGULARIZATION, NORMALIZATION
from .graph import network_indices, BlockLaplacian


class LaplacianSmoothingModel(object):
    r'''A model that uses a graphical representation of a set of related entities
    to estimate a connectedness-aware version of a quality score :math:`s`.

    The `Graph Laplacian <https://en.wikipedia.org/wiki/Laplacian_matrix>`_ is used
    to summarize the connectedness of nodes in the graph.

    Related subgraphs are further organized into neighborhoods, as defined by
    :attr:`neighborhood_walker`'s rules. These groups are expected to trend together,
    sharing a parameter :math:`\tau`. A single vertex in the graph can belong to multiple
    neighborhoods to different extents, having a belongingness weight for each neighborhood
    encoded in a belongingness matrix :math:`\mathbf{A}`.
    '''

    network: CompositionGraph
    neighborhood_walker: NeighborhoodWalker

    belongingness_matrix: np.ndarray
    variance_matrix: np.ndarray
    inverse_variance_matrix: np.ndarray

    block_L: BlockLaplacian
    obs_ix: np.ndarray
    miss_ix: np.ndarray

    S0: np.ndarray
    A0: np.ndarray
    Am: np.ndarray

    C0: List[CompositionGraphNode]

    threshold: float
    regularize: float

    _belongingness_normalization: str

    def __init__(self, network, belongingness_matrix, threshold,
                 regularize=DEFAULT_LAPLACIAN_REGULARIZATION, neighborhood_walker=None,
                 belongingness_normalization=NORMALIZATION, variance_matrix=None,
                 inverse_variance_matrix=None):
        self.network = network

        if neighborhood_walker is None:
            self.neighborhood_walker = NeighborhoodWalker(self.network)
        else:
            self.neighborhood_walker = neighborhood_walker

        self.belongingness_matrix = belongingness_matrix
        self.threshold = threshold

        self.obs_ix, self.miss_ix = network_indices(self.network, self.threshold)
        self.block_L = BlockLaplacian(
            self.network, threshold, regularize=regularize)

        self.S0 = np.array([node.score for node in self.network[self.obs_ix]])
        self.A0 = belongingness_matrix[self.obs_ix, :]
        self.Am = belongingness_matrix[self.miss_ix, :]
        self.C0 = [node for node in self.network[self.obs_ix]]

        self._belongingness_normalization = belongingness_normalization
        if variance_matrix is None:
            variance_matrix = np.eye(len(self.S0))
        self.variance_matrix = variance_matrix
        if inverse_variance_matrix is None:
            inverse_variance_matrix = np.eye(len(self.S0))
        self.inverse_variance_matrix = inverse_variance_matrix

    def __reduce__(self):
        return self.__class__, (
            self.network, self.belongingness_matrix, self.threshold,
            self.block_L.regularize, self.neighborhood_walker,
            self._belongingness_normalization, self.variance_matrix,
            self.inverse_variance_matrix)

    @property
    def L_mm_inv(self) -> np.ndarray:
        return self.block_L.L_mm_inv

    @property
    def L_oo_inv(self) -> np.ndarray:
        return self.block_L.L_oo_inv

    def optimize_observed_scores(self, lmda: float, t0: np.ndarray=0) -> np.ndarray:
        blocks = self.block_L
        # Here, V_inv is the inverse of V, which is the inverse of the variance matrix
        V_inv = self.variance_matrix
        L = lmda * V_inv.dot(blocks["oo"] - blocks["om"].dot(self.L_mm_inv).dot(blocks["mo"]))
        B = np.identity(len(self.S0)) + L
        return np.linalg.inv(B).dot(self.S0 - t0) + t0

    def compute_missing_scores(self, observed_scores: np.ndarray, t0=0., tm=0.) -> np.ndarray:
        blocks = self.block_L
        return -linalg.inv(blocks['mm']).dot(blocks['mo']).dot(observed_scores - t0) + tm

    def compute_projection_matrix(self, lmbda) -> np.ndarray:
        A = np.eye(self.L_oo_inv.shape[0]) + self.L_oo_inv * (1. / lmbda)
        H = np.linalg.pinv(A)
        return H

    def compute_press(self, observed: np.ndarray, updated: np.ndarray, projection_matrix: np.ndarray) -> np.ndarray:
        press = np.sum(((observed - updated) / (
            1 - (np.diag(projection_matrix) - np.finfo(float).eps))) ** 2) / len(observed)
        return press

    def estimate_tau_from_S0(self, rho: float, lmda: float, sigma2: float=1.0) -> np.ndarray:
        X = ((rho / sigma2) * self.variance_matrix) + (
            (1. / (lmda * sigma2)) * self.L_oo_inv) + self.A0.dot(self.A0.T)
        X = np.linalg.pinv(X)
        return self.A0.T.dot(X).dot(self.S0)

    def estimate_phi_given_S_parameters(self, lmbda: float, tau: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        r"""Estimate the parameters of the conditional distribution :math:`\phi|s`
        according to [1]_.

        Notes
        -----
        According to [1]_, given
        .. math::

                y | \theta_1 \sim & \mathcal{N}\left(\mathbf{A_1}\theta_1, \mathbf{C_1}\right) \\

                \theta_1 \sim & \mathcal{N}\left(\mathbf{A_2}\theta_2, \mathbf{C_2}\right) \\

                \theta_1 | y \sim & \mathcal{N}\left(\mathbf{B}b, \mathbf{B}\right) \\

                \mathbf{B^{-1}} = & \mathbf{C_2}^{-1} + \mathbf{A_1}^t\mathbf{C_1}^{-1}\mathbf{A_1} \\
                b = & \mathbf{A_1}^t\mathbf{C_1}^{-1}y + \mathbf{C_2}^{-1}\mathbf{A_2}\theta_2

        In terms of the network smoothing model, this translates to

        .. math::

                \mathbf{B^{-1}} = & \lambda\mathbf{L} + \tilde{A}^t\Sigma^{-1}\tilde{A}\\
                                = & \lambda \mathbf{L} + \begin{bmatrix} \Sigma_{oo}^{-1} & 0 \\ 0 & 0 \end{bmatrix} \\

                b = & \tilde{A}^t\Sigma^{-1}s + \lambda\mathbf{L}\mathbf{A}\tau \\
                  = & \begin{bmatrix}\Sigma_{oo}^{-1}s \\ 0\end{bmatrix} + \lambda\begin{bmatrix}L_{oo}A\tau_o +
                      L_{om}A\tau_m \\L_{mo}A\tau_o + L_{mm}A\tau_m\end{bmatrix}\\\\

        Parameters
        ----------
        lmbda : float
            The smoothing factor
        tau : np.ndarray
            The mean of :math:`\phi`, :math:`\mathbf{A}\tau`

        Returns
        -------
        Bb: np.ndarray
            The mean vector of :math:`\phi|s`
        B: np.ndarray:
            The variance matrix of :math:`\phi|s`

        References
        ----------
        .. [1] Lindley, D. V., & Smith, A. F. M. (1972). Bayes Estimates for the Linear Model.
               Royal Statistical Society, 34(1), 1–41.
        """
        B_inv = lmbda * self.block_L.matrix
        B_inv[self.obs_ix, self.obs_ix] += np.diag(self.inverse_variance_matrix)
        B = linalg.inv(B_inv)

        Alts = np.zeros(len(self.network))
        Alts[self.obs_ix] = self.S0.dot(self.inverse_variance_matrix)
        lambda_Lw_Atau = lmbda * self.block_L.matrix.dot(self.normalized_belongingness_matrix.dot(tau))
        b = Alts + lambda_Lw_Atau
        phi_given_s = np.dot(B, b)
        return phi_given_s, B

    def build_belongingness_matrix(self) -> np.ndarray:
        belongingness_matrix = self.neighborhood_walker.build_belongingness_matrix()
        return belongingness_matrix

    def get_belongingness_patch(self) -> np.ndarray:
        updated_belongingness = BelongingnessMatrixPatcher.patch(self)
        updated_belongingness = ProportionMatrixNormalization.normalize(
            updated_belongingness, self._belongingness_normalization)
        return updated_belongingness

    def apply_belongingness_patch(self):
        updated_belongingness = self.get_belongingness_patch()
        self.belongingness_matrix = updated_belongingness
        self.A0 = self.belongingness_matrix[self.obs_ix, :]


class SmoothedScoreSample(object):
    def __init__(self, model, lmbda, tau, size=10000, random_state=None):
        if random_state is None:
            random_state = np.random.RandomState()
        self.model = model
        self.lmbda = lmbda
        self.tau = tau
        self.size = size
        self.Bb, self.B = self.model.estimate_phi_given_S_parameters(lmbda, tau)
        self.B_inv = np.linalg.pinv(self.B)
        self.random_state = random_state
        self.samples = np.array([])
        self.resample()

    def resample(self):
        self.samples = self.random_state.multivariate_normal(self.Bb, self.B, size=self.size)
        # alternatively, sample unit variance scaled by the cholesky decomposition of the covariance
        # matrix to get the sample variance
        # C = linalg.cholesky(self.B)
        # self.samples = np.array(
        #                   [self.Bb + C.T.dot(self.random_state.normal(size=C.shape[0])) for i in range(self.size)])
        return self

    def _get_index(self, glycan_composition):
        node = self.model.network[glycan_composition]
        index = node.index
        return index

    def _score_index(self, index):
        return self.Bb[index]

    def _gaussian_pdf(self, x, mu, sigma, sigma_inv):
        d = (x - mu)
        e = np.exp((-0.5 * (d) * sigma_inv * d))
        return e / np.sqrt(2 * np.pi * sigma)

    def score(self, glycan_composition):
        index = self._get_index(glycan_composition)
        return self._score_index(index)

    def plot_distribution(self, glycan_composition, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        index = self._get_index(glycan_composition)
        node = self.model.network[index]
        s = self.samples[:, index]
        bins = np.arange(0, self.Bb[index] * 2, .1)
        ax.hist(s, bins=bins, alpha=0.5, density=True, label="%s (%f)" % (node, self.score(glycan_composition)))
        ax.set_xlabel(r"Resampled $\phi$", fontsize=24)
        ax.set_ylabel("Density", fontsize=24)
        return ax


class ProportionMatrixNormalization(object):
    def __init__(self, matrix):
        self.matrix = np.array(matrix)

    def normalize_columns(self):
        self.matrix = self.matrix / self.matrix.sum(axis=0)

    def normalize_rows(self):
        normalizer = self.matrix.sum(axis=1).reshape((-1, 1))
        normalizer[normalizer == 0] = 1e-6
        self.matrix = self.matrix / normalizer

    def normalize_columns_and_rows(self):
        self.normalize_columns()
        self.clean()
        self.normalize_rows()

    def normalize_columns_scaled(self, scaler=2.0):
        self.normalize_columns()
        self.matrix = self.matrix * scaler

    def clean(self):
        self.matrix[np.isnan(self.matrix)] = 0.0

    def __array__(self):
        return self.matrix

    @classmethod
    def normalize(cls, matrix, method='colrow'):
        self = cls(matrix)
        if method == 'col':
            self.normalize_columns()
        elif method == 'row':
            self.normalize_rows()
        elif method == 'colrow':
            self.normalize_columns_and_rows()
        elif method is not None and method.startswith("col") and method[3:4].isdigit():
            scale = float(method[3:])
            self.normalize_columns_scaled(scale)
        elif method == 'none' or method is None:
            pass
        else:
            raise ValueError("Unknown Normalization Method %r" % method)
        self.clean()
        return self.matrix


class GroupBelongingnessMatrix(object):
    _node_like = (basestring, CompositionGraphNode,
                  GlycanComposition, GlycanCompositionProxy)

    @classmethod
    def from_model(cls, model, normalized=True):
        if normalized:
            mat = model.normalized_belongingness_matrix
        else:
            mat = model.belongingness_matrix
        groups = model.neighborhood_walker.neighborhood_names()
        members = model.network.nodes
        return cls(mat, groups, members)

    def __init__(self, belongingness_matrix, groups, members):
        self.belongingness_matrix = np.array(belongingness_matrix)
        self.groups = [str(x) for x in groups]
        self.members = [str(x) for x in members]
        self._column_indices = OrderedDict([(k, i) for i, k in enumerate(self.groups)])
        self._member_indices = OrderedDict([(k, i) for i, k in enumerate(self.members)])

    def _column_indices_by_name(self, names):
        if isinstance(names, basestring):
            names = [names]
        indices = [self._column_indices[n] for n in names]
        return indices

    def _member_indices_by_name(self, names):
        if isinstance(names, self._node_like):
            names = [names]
        indices = [self._member_indices[n] for n in names]
        return indices

    def _coerce_member(self, names):
        if isinstance(names, self._node_like):
            names = [names]
        return names

    def _coerce_column(self, names):
        if isinstance(names, basestring):
            names = [names]
        return names

    def get(self, rows=None, cols=None):
        matrix = self.belongingness_matrix
        if rows is not None:
            rows = self._coerce_member(rows)
            row_ix = self._member_indices_by_name(rows)
            matrix = matrix[row_ix, :]
        else:
            rows = self.members
        if cols is not None:
            cols = self._coerce_column(cols)
            col_ix = self._column_indices_by_name(cols)
            matrix = matrix[:, col_ix]
        else:
            cols = self.groups
        return self.__class__(matrix, cols, rows)

    def getindex(self, rows=None, cols=None):
        if rows is not None:
            rows = self._coerce_member(rows)
            row_ix = self._member_indices_by_name(rows)
        else:
            row_ix = None
        if cols is not None:
            cols = self._coerce_column(cols)
            col_ix = self._column_indices_by_name(cols)
        else:
            col_ix = None
        return row_ix, col_ix

    def __getitem__(self, ij):
        return self.belongingness_matrix[ij]

    def __setitem__(self, ij, val):
        self.belongingness_matrix[ij] = val

    def __array__(self):
        return np.array(self.belongingness_matrix)

    def __repr__(self):
        column_label_width = max(map(len, self.groups))
        row_label_width = max(map(len, self.members))
        rows = []
        top_row = [' ' * row_label_width]
        for col in self.groups:
            top_row.append(col.center(column_label_width))
        rows.append('|'.join(top_row))
        for i, member in enumerate(self.members):
            row = [member.ljust(row_label_width)]
            vals = self.belongingness_matrix[i, :]
            for val in vals:
                row.append(("%0.3f" % val).center(column_label_width))
            rows.append("|".join(row))
        return '\n'.join(rows)


MatrixEditIndex = namedtuple("MatrixEditIndex", ("row_index", "col_index", "action"))
MatrixEditInstruction = namedtuple("MatrixEditInstruction", ("composition", "neighborhood", "action"))


class BelongingnessMatrixPatcher(object):
    def __init__(self, model):
        self.model = model
        self.A0 = model.A0
        self.belongingness_matrix = GroupBelongingnessMatrix.from_model(
            model, normalized=False)

    def find_singleton_neighborhoods(self):
        n_cols = self.A0.shape[1]
        edits = []
        for i in range(n_cols):
            # We have a neighborhood with only one member
            col = self.A0[:, i]
            mask = col > 0
            if mask.sum() == 1:
                j = np.argmax(mask)
                edit = MatrixEditIndex(j, i, 'delete')
                edits.append(edit)
        return edits

    def transform_index_to_key(self, edits):
        neighborhood_names = self.model.neighborhood_walker.neighborhood_names()
        out = []
        for edit in edits:
            key = str(self.model.C0[edit.row_index])
            neighborhood = neighborhood_names[edit.col_index]
            out.append(MatrixEditInstruction(key, neighborhood, edit.action))
        return out

    def patch_belongingness_matrix(self, edits):
        gbm = self.belongingness_matrix
        instructions = self.transform_index_to_key(edits)
        for instruction in instructions:
            ij = gbm.getindex(instruction.composition, instruction.neighborhood)
            if instruction.action == 'delete':
                gbm[ij] = 0
        return gbm

    @classmethod
    def patch(cls, model):
        inst = cls(model)
        targets = inst.find_singleton_neighborhoods()
        out = inst.patch_belongingness_matrix(targets)
        return np.array(out)
